from logging_helper import setup_logging

from ._cache import get_file_cache
from ._compat import urlsplit

_lp_client = None
logger = setup_logging()


def _get_lp():
    from lplight.client import LaunchpadClient
    import lplight.client

    web_session = get_file_cache("launchpad")
    lplight.client.requests = web_session

    global _lp_client

    if _lp_client is None:
        _lp_client = LaunchpadClient()
    return _lp_client


def _launchpad_bug(bug_id):
    lp_client = _get_lp()
    try:
        r = lp_client.get_bug_tasks(bug_id)
    except Exception as e:  # pragma: no cover
        logger.error("_launchpad_bug: {}".format(e))
        return

    if not r:  # pragma: no cover
        return
    data = r.json()
    project1 = data["entries"][0]["bug_target_name"]
    return _launchpad("https://launchpad.net/{}".format(project1))


def _launchpad(url):
    p = urlsplit(url)
    subdomain = p.netloc.split(".")[0]
    if subdomain == "launchpad":
        subdomain = None

    path = p.path[1:]
    if not path:  # pragma: no cover
        return
    name, _, other = path.partition("/")

    if not name:  # pragma: no cover
        return

    if name == "bugs":
        bug_id = other
        return _launchpad_bug(bug_id)
    elif name == "api":  # pragma: no cover
        path2, _, other = other.partition("/")
        path3, _, other = other.partition("/")
        name = path3
    elif name[0] == "~":
        if subdomain in ["bazaar", "code"]:
            path2, _, other = other.partition("/")
            path3, _, other = other.partition("/")
            name += "/" + path2 + "/" + path3
        else:
            return

    if not name:  # pragma: no cover
        raise RuntimeError("could not extract name from {}".format(url))

    lp_client = _get_lp()
    try:
        r = lp_client.get_project(name)
    except Exception as e:
        logger.error("_launchpad: {}".format(e))
        return
    if not r:
        return
    data = r.json()
    if not data.get("is_valid", True):  # pragma: no cover
        return

    check_url = data.get("homepage_url")
    logger.debug("lp homepage: {}".format(check_url))
    if not check_url:
        bzr_identity = data.get("bzr_identity")
        if bzr_identity and bzr_identity.startswith("lp:"):
            logger.debug("lp bzr id: {}".format(bzr_identity))
            return _launchpad("https://launchpad.net/{}".format(bzr_identity[3:]))

        check_url = data["web_link"]
        logger.debug("lp weblink: {}".format(check_url))

    if check_url != url:
        from ._scm_url_cleaner import SCMURLCleaner

        subcleaner = SCMURLCleaner()
        if "launchpad.net" not in check_url:
            logger.info("checking url {}".format(check_url))
            redirect = subcleaner.get_root(check_url)
            logger.info("checking url {} = ".format(redirect))
            if redirect:
                return redirect

    if "pypi.org" in url or "pypi.python.org" in url:
        return

    return "https://launchpad.net/{}".format(name)



import unittest

from unittest_expander import expand, foreach

from pypidb._compat import PY2
from pypidb._db import _fetch_mapping
from pypidb._github import GitHubAPIMessage, check_repo
from pypidb._pypi import InvalidPackage
from pypidb._similarity import _compute_similarity, normalize
from tests.data import (
    exact,
    exact_fetched,
    exact_metadata,
    invalid,
    mismatch,
    missing_repos,
    name_mismatch_fetched,
    name_mismatch_metadata,
)
from tests.utils import _TestBase, normalise_list, web_session

expected = {}
expected.update(exact)
expected.update(mismatch)

missing_repos = normalise_list(missing_repos)


class _ExplicitBase(object):
    def _check_result(self, name, url):
        pass

    def _test_package(self, name):
        try:
            url = self._get_scm(name)
        except InvalidPackage:
            self.assertIn(name, invalid)
            raise unittest.SkipTest("{} is an invalid package".format(name))
        except unittest.SkipTest:
            raise
        except Exception as e:
            self.assertIn(name, self.expected_failures, e)

        if name in self.expected_failures:
            return
        self.assertIsNotNone(url)
        self.assertNotIn(name, invalid)

        expected = self.expected[name]

        self.assertInsensitiveEqual(url, expected)
        self._check_result(name, url)

        return url


@expand
class TestExactFromJson(_TestBase):

    expected = expected

    def _check_result(self, name, url):
        normalised_name = normalize(name)

        r = None
        if normalised_name in missing_repos:
            pass
        elif url.startswith("https://github.com/"):
            slug = url[len("https://github.com/") :]
            rv = self._check_github_repo(slug)

            self.assertTrue(rv)

            try:
                rv = self._check_github_setuppy(slug, normalised_name)
            except GitHubAPIMessage as e:
                raise unittest.SkipTest(str(e))

            if rv is False:
                return
            self.assertTrue(rv)

        else:
            r = web_session.get(url, allow_redirects=False)

        if r is not None:
            r.raise_for_status()
            self.assertEqual(r.url, url)
            location = r.headers.get("location")
            if location:
                self.assertIn(r.status_code, [301, 302])
                location = location.replace(
                    "code.google.com/archive/p/", "code.google.com/p/"
                )
                self.assertIn(location, [url, url + "/"])
            else:
                self.assertEqual(r.status_code, 200)

    @foreach(exact_metadata.keys())
    def test_package(self, name):
        expected = self.expected[name]

        url = self._get_scm(name)

        self.assertIsNotNone(url)
        self.assertInsensitiveEqual(url, expected)

        normalised_name = normalize(name)
        fetch_list = _fetch_mapping[normalised_name]
        self.assertFalse(fetch_list)

        self._check_result(name, url)

        if PY2:
            return

        if isinstance(expected, str):
            self.assertLess(
                _compute_similarity(name, expected),
                0.05,
                "{} - {} should be moved to name mismatches".format(name, expected),
            )
        else:
            for i in expected:
                self.assertLess(
                    _compute_similarity(name, i),
                    0.05,
                    "{} - {} should be moved to name mismatches".format(name, expected),
                )


@expand
class TestExactFetched(_TestBase):

    expected = expected

    @foreach(exact_fetched.keys())
    def test_package(self, name):
        expected = self.expected[name]

        url = self._get_scm(name)

        self.assertIsNotNone(url)
        self.assertInsensitiveEqual(url, expected)

        normalised_name = normalize(name)
        fetch_list = _fetch_mapping[normalised_name]
        self.assertTrue(fetch_list)

        if normalised_name in missing_repos:
            pass
        elif url.startswith("https://github.com/"):
            slug = url[len("https://github.com/") :]
            rv = self._check_github_repo(slug)

            self.assertTrue(rv)

            try:
                rv = self._check_github_setuppy(slug, normalised_name)
            except GitHubAPIMessage as e:
                raise unittest.SkipTest(str(e))

            if rv is False:
                return
            self.assertTrue(rv)

        else:
            r = web_session.get(url)
            r.raise_for_status()

        if PY2:
            return

        if isinstance(expected, str):
            self.assertLess(
                _compute_similarity(name, expected),
                0.05,
                "{} - {} should be moved to name mismatches".format(name, expected),
            )
        else:
            for i in expected:
                self.assertLess(
                    _compute_similarity(name, i),
                    0.05,
                    "{} - {} should be moved to name mismatches".format(name, expected),
                )


@expand
class TestMismatchFromJson(_TestBase):

    expected = expected
    names = mismatch
    expected_failures = []

    @foreach(name_mismatch_metadata.keys())
    def test_package(self, name):
        expected = self.expected[name]

        try:
            url = self._get_scm(name)
        except unittest.SkipTest:
            raise
        except Exception:
            if name in self.expected_failures:
                return
            raise

        if name in self.expected_failures:
            return

        self.assertIsNotNone(url)
        self.assertInsensitiveEqual(url, expected)
        if isinstance(expected, str):
            self.assertIn("/", expected, "{} should be {}".format(expected, url))

        normalised_name = normalize(name)
        fetch_list = _fetch_mapping[normalised_name]
        self.assertFalse(fetch_list)

        if normalised_name in missing_repos:
            pass
        elif url.startswith("https://github.com/"):
            slug = url[len("https://github.com/") :]
            rv = self._check_github_repo(slug)

            self.assertTrue(rv)

            try:
                rv = self._check_github_setuppy(slug, normalised_name)
            except GitHubAPIMessage as e:
                raise unittest.SkipTest(str(e))

            if rv is False:
                return
            self.assertTrue(rv)

        else:
            r = web_session.get(url)
            r.raise_for_status()

        if PY2:
            return

        if isinstance(expected, str):
            self.assertGreater(_compute_similarity(name, expected), 0.05)
        else:
            highest = 0
            for i in expected:
                val = _compute_similarity(name, i)
                highest = max(highest, val)
            self.assertGreater(highest, 0.05)


@expand
class TestMismatchFetched(_ExplicitBase, _TestBase):

    expected = expected
    names = mismatch
    expected_failures = ["marionette-driver"]

    @foreach(name_mismatch_fetched.keys())
    def test_package(self, name):
        expected = self.expected[name]

        url = self._test_package(name)

        if name in self.expected_failures:
            return

        self.assertIsNotNone(url)
        self.assertInsensitiveEqual(url, expected)
        if isinstance(expected, str):
            self.assertIn("/", expected, "{} should be {}".format(expected, url))

        normalised_name = normalize(name)
        fetch_list = _fetch_mapping[normalised_name]

        self.assertTrue(fetch_list)

        if normalised_name in missing_repos:
            pass
        elif url.startswith("https://github.com/"):
            slug = url[len("https://github.com/") :]
            rv = self._check_github_repo(slug)

            self.assertTrue(rv)

            try:
                rv = self._check_github_setuppy(slug, normalised_name)
            except GitHubAPIMessage as e:
                raise unittest.SkipTest(str(e))

            if rv is False:
                return
            self.assertTrue(rv)

        elif url == "https://wiki.mozilla.org/Auto-tools/Projects/Mozbase":
            # Fetching is a bit slow, and failures for moz* are very repetitive
            pass
        else:
            r = web_session.get(url)
            r.raise_for_status()

        if PY2:
            return

        if isinstance(expected, str):
            self.assertGreater(_compute_similarity(name, expected), 0.05)
        else:
            highest = 0
            for i in expected:
                val = _compute_similarity(name, i)
                highest = max(highest, val)
            self.assertGreater(highest, 0.05)



import unittest

from unittest_expander import expand, foreach

from tests.data import bad_metadata, expected
from tests.utils import _stdlib, _TestBase, normalise_list

try:
    from tests.datasets import _intersect_stdlib, get_top_packages
except (SyntaxError, ImportError):
    raise unittest.SkipTest("Disabled on Python 2")


expected = normalise_list(expected)
bad_metadata = normalise_list(bad_metadata)

failures = []

repeat_expected = False


def _fetch_names(kind="top4kyear", xstatic=False, ignore_failures=True):
    for name in get_top_packages(kind):
        if name in _stdlib:
            continue

        if xstatic:
            if not name.startswith("xstatic-"):
                continue
        else:
            if name.startswith("xstatic-"):
                continue

        if not repeat_expected and name.lower() in expected:
            continue

        if name.lower() in [i.lower() for i in bad_metadata]:
            continue

        if ignore_failures and name in [i.lower() for i in failures]:
            continue

        yield name


_all = list(_fetch_names())
_month = list(name for name in _fetch_names("top4kmonth") if name not in _all)


@expand
class TestTopStdlib(_TestBase):

    names = list(_intersect_stdlib())
    # TestStdlibNames allows missing metadata;
    # this top class does not, as any highly used
    # bit of software should be in a VCS somewhere
    # and be findable, otherwise many eyes doesnt work.
    expected_failures = [
        "ast",
        "email",
        "enum",
        "faulthandler",
        "functools",
        "html",
        "pprint",
        "resource",
        "typing",
        "uuid",
    ]

    @foreach(names)
    def test_package(self, name):
        self._test_names([name])


@expand
class TestTop360(_TestBase):
    # 360 to match https://github.com/hugovk/drop-python/

    names = _all[:360]
    expected_failures = _TestBase.expected_failures + [
        "newrelic",
        "nvidia-ml-py3",  # new, no urls, todo?
        "ordereddict",  # old, no urls
        "termcolor",  # no vcs
        "widgetsnbextension",  # subproject of ipywidgets
    ]

    @foreach(names)
    def test_package(self, name):
        self._test_names([name])


@expand
class TestTopXStatic(_TestBase):

    names = set(_fetch_names(xstatic=True, kind="top4kyear")) | set(
        _fetch_names(xstatic=True, kind="top4kmonth")
    )

    @foreach(names)
    def test_package(self, name):
        self._test_names([name])


@expand
class TestTopFirstThousandTail(_TestBase):

    names = _all[360:1000]
    expected_failures = _TestBase.expected_failures + [
        "google-cloud-dataflow",  # no urls
        "intel-openmp",
        "marshmallow-enum",  # no urls
        "mysql-connector-python-rf",
        "mysql-connector",
        "mysql-connector-python",
        "nvidia-ml-py",
        "requestsexceptions",  # c.f. test_openstack.py
        "newrelic",  # no urls
        "spotinst-agent",  # https://github.com/spotinst/spotinst-sdk-python/issues/65
    ]

    @foreach(names)
    def test_package(self, name):
        self._test_names([name])


@expand
class TestTopSecondThousand(_TestBase):

    names = _all[1000:2000]
    expected_failures = _TestBase.expected_failures + ["oset"]

    @foreach(names)
    def test_package(self, name):
        self._test_names([name])


@expand
class TestTopTail(_TestBase):

    names = _all[2000:]
    expected_failures = [
        "bindep",
        "clearbit",
        "comet-ml",
        "dbus-python",
        "dm.xmlsec-binding",
        "pycuda",  # repo is auth protected
        "robinhood-aiokafka",
        "scons",
        "tableauhyperapi",
        "tensorboard-plugin-wit",
    ]

    @foreach(names)
    def test_package(self, name):
        self._test_names([name])


@expand
class TestTop30Days(_TestBase):

    names = _month
    expected_failures = _TestBase.expected_failures + [
        "awshelpers",
        "azureml-mlflow",  # no urls found
        "comet-ml",
        "edx-tincan-py35",
        "clearbit",
        "kivy-deps-sdl2",  # no suitable metadata
        "mandrill-37",
        "mxnet-cu90mkl",
        "netapp-lib",
        "pretty-bad-protocol",  # todo
        "replit",
        "strsim",  # old version of strsimpy
        "testinfiniteloop",
        "tensorflow-addons",
        "watson-machine-learning-client-v4",
    ]

    @foreach(names)
    def test_package(self, name):
        self._test_names([name])



import unittest

from stdlib_list import stdlib_list

from requests.exceptions import HTTPError

from pypidb._cache import get_file_cache, get_timeout
from pypidb._compat import PY2
from pypidb._db import Database, multipackage_repos
from pypidb._github import check_repo as check_github_repo, get_repo_setuppy
from pypidb._github import GitHubAPIMessage
from pypidb._pypi import IncompletePackageMetadata, InvalidPackage
from pypidb._rules import DefaultRule, Rule, rules
from pypidb._similarity import _compute_similarity, normalize
from tests.data import (
    mismatch,
    missing_repos,
    setuppy_mismatches,
    setuppy_mismatches_xstatic,
    wrong_result,
)

_stdlib_all = stdlib_list()
_stdlib = set(i.split(".")[0] for i in _stdlib_all)

USE_SEPARATE_INSTANCE = (
    True
)  # This can only be disabled if only one set of tests are run

web_session = get_file_cache("web")


def normalise_list(in_list):
    return [normalize(name) for name in in_list]


_global_converter = _global_db = Database(
    website_timeout=(15, 30), store_fetch_list=True
)


class _TestBase(unittest.TestCase):

    names = []

    expected_failures = ["sklearn", "bs4"]  # joejob/redirect  # "
    _run_expected_failures = True
    _ignore_invalid = False
    _allow_missing = False
    _debug_no_urls = False
    _allow_none = False
    _debug_multi_urls = False  # True

    def setUp(self):
        if USE_SEPARATE_INSTANCE:
            self.converter = Database(
                website_timeout=(15, 30), store_fetch_list=True
            )._converter
        else:
            self.converter = _global_converter

    def tearDown(self):
        pass

    def assertInsensitiveEqual(self, s, s2):
        if not s and self._allow_none:
            return

        self.assertIsNotNone(s)
        if isinstance(s2, str):
            if "/" in s2:
                self.assertEqual(s.lower(), s2.lower())
            else:
                self.assertIn("/" + s2.lower(), s.lower())
        else:
            self.assertIn(s.lower(), [i.lower() for i in s2])
            if self._debug_multi_urls:
                self.assertIsInstance(s2, str, "{} matches {!r}".format(s, s2))

    def assertRaisesNoUrls(self, name):
        try:
            url = self.converter.get_vcs(name)
        except IncompletePackageMetadata:
            if self._debug_no_urls:
                self.assertTrue(False)
            return
        self.assertIsNotNone(url)
        self.assertIsNone(url)

    def assertNoUrls(self, s):
        if self._debug_no_urls:
            self.assertIsNotNone(s)
        else:
            self.assertIsNone(s)

    def _get_scm(self, name):
        try:
            url = self.converter.get_vcs(name)
        except HTTPError as e:
            if "403" in str(e):
                raise unittest.SkipTest(str(e))
            raise
        except GitHubAPIMessage as e:
            raise unittest.SkipTest(str(e))
        return url

    def _check_github_repo(self, slug):
        try:
            rv = check_github_repo(slug)
        except GitHubAPIMessage as e:
            raise unittest.SkipTest(str(e))
        except HTTPError as e:
            if "403" in str(e):
                raise unittest.SkipTest(str(e))
            raise
        return rv

    def _check_github_setuppy(self, slug, rule):
        if not isinstance(rule, Rule):
            normalised_name = rule
            rule = rules.get(normalised_name)
            if not rule:
                rule = DefaultRule(normalised_name)

        for setuppy_rule in setuppy_mismatches:
            setuppy_rule = normalize(setuppy_rule.strip("$"))
            if rule.key.startswith(setuppy_rule):
                return False

        matches = [rule.key]
        if normalised_name in setuppy_mismatches_xstatic:
            matches = ["xstatic-pkg", normalised_name[8:]]

        filenames = None
        try:
            filenames = [rule.repo_filename]
        except AttributeError:
            pass

        try:
            rv = get_repo_setuppy(slug, matches, filenames)
        except GitHubAPIMessage as e:
            raise unittest.SkipTest(str(e))
        except HTTPError as e:
            if "403" in str(e):
                raise unittest.SkipTest(str(e))
            raise

        if rv is not None:
            match_results = [match in normalize(rv) for match in matches]
            assert all(match_results), "{}: {} not in:\n{}".format(
                rule.key, matches, rv
            )

        return rv

    def _test_names(self, names, ignore_not_found=False, ignore_setuppy=False):
        assert len(names) == 1
        expected_failures = [normalize(i) for i in self.expected_failures]

        name = names[0]
        normalised_name = normalize(name)

        try:
            url = self._get_scm(name)
        except unittest.SkipTest:
            raise
        except InvalidPackage:
            if (
                normalised_name in expected_failures
                or self._ignore_invalid
                or ignore_not_found
            ):
                return

            raise
        except IncompletePackageMetadata:
            if PY2:
                return
            if (
                normalised_name in expected_failures
                or self._ignore_invalid
                or self._allow_missing
                or ignore_not_found
            ):
                return

            raise
        except Exception:
            if PY2 or normalised_name in expected_failures:
                return

            raise

        if not url:
            if (
                normalised_name in expected_failures
                or self._allow_missing
                or ignore_not_found
            ):
                return

            assert False, name

        if normalised_name in missing_repos or normalised_name in expected_failures:
            return

        if name in wrong_result:
            return

        distance = _compute_similarity(name, url)

        if distance < 0.1:
            pass
        else:
            normalised_name = normalize(name)
            mismatch_urls = [
                value
                for key, value in mismatch.items()
                if normalize(key) == normalised_name
            ]
            if url.lower() in [
                "https://github.com/azure/azure-sdk-for-python",
                "https://github.com/azure/azure-cli",
            ]:
                return
            elif mismatch_urls:
                if (
                    url not in multipackage_repos
                    and url.lower() not in multipackage_repos
                ):
                    assert len(mismatch_urls) == 1, "{} matched {}".format(
                        name, mismatch_urls
                    )
                expected_url = mismatch_urls[0]
                self.assertInsensitiveEqual(url, expected_url)
            else:
                if True:
                    for rule in setuppy_mismatches:
                        if rule.endswith("-") and normalised_name.startswith(rule):
                            break
                    else:
                        raise Exception(
                            '"{}": "{}" should be in mismatches'.format(name, url)
                        )

        if url.startswith("https://github.com/"):

            slug = url[len("https://github.com/") :]
            try:
                rv = self._check_github_repo(slug)
            except unittest.SkipTest:
                raise
            except Exception:
                rv = None
            if not rv:
                if PY2 or name in self.expected_failures:
                    return
                elif name in wrong_result:
                    return
                elif ignore_setuppy:
                    return
                assert False, slug

            self._check_github_setuppy(slug, normalised_name)

        else:
            r = web_session.get(url, timeout=get_timeout(url))
            r.raise_for_status()



import unittest

from r2c_isg.apis import api_map
from r2c_isg.loaders import Loader
from r2c_isg.loaders.web import webloader_map
from r2c_isg.structures import Dataset
from r2c_isg.structures.projects import PypiProject
from r2c_isg.structures.projects import project_map as registry_map
from tqdm import tqdm

from pypidb._compat import PY2
from tests.utils import _stdlib_all

if PY2:
    raise unittest.SkipTest("datasets not supported on Python 2")

R2C_WEB_CACHE = "tests/.datasets_requests_cache"


class PortingdbLoader(Loader):

    _pypi_name_mapping = {
        "django-database-url": "dj-database-url",
        "django-email-url": "dj-email-url",
        "django-search-url": "dj-search-url",
        "dns": "dnspython",
        "epi": "entry_point_inspector",
        "github3py": "github3.py",
        "glusterfs-api": "gfapi",
        # 'hwdata': 'pciutils', according to https://github.com/xsuchy/python-hwdata/blob/master/setup.py , but that is also not pypi name
        "jupyter-polymake": "jupyter-kernel-polymake",
        "libcloud": "apache-libcloud",
        "novaclient-os-networks": "os_networksv2_python_novaclient_ext",
        "novaclient-os-virtual-interfaces": "os_virtual_interfacesv2_python_novaclient_ext",
        "podman-api": "podman",
        "pycolumnize": "columnize",
        "pyhunspell": "hunspell",
        "pyobd": "obd",
        "pyosmium": "osmium",
        "pywt": "PyWavelets",
        "rtslib": "rtslib-fb",
        "sphinx-theme-alabaster": "alabaster",
        "sphinx-theme-py3doc-enhanced": "sphinx-py3doc-enhanced-theme",
        "uri-templates": "uritemplate",
        "webpy": "web.py",
        "XStatic-termjs": "XStatic-term.js",
        "zmq": "pyzmq",
    }

    _not_pypi = [
        "azure-sdk",
        "basemap-data",
        "blivet",
        "caja",
        "collectd_systemd",  # pypi_name is wrong
        "conda-package-handling",
        "configshell",
        "coverage-test-runner",
        "cpio",
        "distutils-extra",  # http://launchpad.net/python-distutils-extra/
        "docs",
        "evic",  # pypi_name is wrong
        "firkin",
        "gradunwarp",
        "gstreamer1",
        "gtkextra",
        "gzipstream",
        "hwdata",  # name should be 'pciutils' according to https://github.com/xsuchy/python-hwdata/blob/master/setup.py , but that is also not pypi name
        "iep",  # pypi URL wrong
        "ipgetter",  # ipgetter2 repo https://github.com/phoemur/ipgetter is missing
        "linux-procfs",  # http://userweb.kernel.org/python-linux-procfs (also mirrored on github)
        "lxc",
        "olpcgames",
        "phyghtmap",
        "pip-epel",
        "postgresql",
        "pyatspi",
        "pybox2d",  # box2d is different https://code.google.com/p/pybox2d
        "pycanberra",
        "pycxx",  # http://cxx.sourceforge.net/
        "pyev",  # setup.py checked; it is pyev
        "pyflowtools",
        "pygiftiio",
        "pygobject2",
        "pygobject3",
        "pygtk2",
        "pyhoca-cli",
        "pyhoca-gui",
        "pyoptical",
        "pyotherside",
        "pyrpmmd",
        "pysvn",
        "pythia8",
        "pyxtrlock",
        "pyzy",
        "rpmautospec",
        "rpm-generators",
        "schedutils",
        "simpleline",
        "smartcols",
        "socksipychain",
        "ttystatus",
        "typeshed",  # Removed from PyPI
        "uranium-lulzbot",  # uranium
        "vevents",
        "virtkey",  # https://launchpad.net/virtkey/
        "visionegg-quest",  # visionegg
        "x2go",
        "xapp",
    ]

    @staticmethod
    def get_pypi_name(package_name):
        if package_name.endswith("rpm-macros"):
            return

        if package_name.startswith("python-"):
            name = package_name[7:]
        elif package_name.startswith("python2-") or package_name.startswith("python3-"):
            name = package_name[8:]
        elif package_name.startswith("py"):
            name = package_name
        else:
            return

        # multiple packages, per version
        if name[-1].isdigit():
            if package_name in ["python26", "python27", "python3"]:
                return
            if package_name[:-1] == "python3":
                return
            if name.startswith("tornado"):
                return
            if name.startswith("pytest"):
                return
            if name.startswith("eric"):
                return
            if name.startswith("Django"):
                return
            if name.startswith("markdown"):
                return
            if name.startswith("prompt_toolkit"):
                return
            if name.startswith("wxpython"):
                return

        if name in PortingdbLoader._not_pypi:
            return

        if name in [
            "barbicanclient",
            "box",
            "can",
            "cinderclient",
            "congressclient",
            "daemon",
            "dateutil",
            "dbusmock",
            "debian",
            "debianbts",
            "designateclient",
            "editor",
            "engineio",
            "fedora",
            "gammu",
            "gilt",
            "glanceclient",
            "heatclient",
            "hglib",
            "hpilo",
            "iptables",
            "ironicclient",
            "k8sclient",
            "Levenshtein",
            "libdiscid",
            "ly",
            "lzo",
            "magic",
            "magnumclient",
            "manilaclient",
            "memcached",  # also python3-memcached
            "mistralclient",
            "mpd",
            "mpd2",
            "multilib",
            "mystrom",
            "networkmanager",
            "neutronclient",
            "novaclient",
            "nss",
            "octaviaclient",
            "opendata-transport",
            "openid",
            "openid-cla",
            "openid-teams",
            "pam",
            "poppler-qt4",
            "poppler-qt5",
            "qt5",
            "registry",
            "rsdclient",
            "saharaclient",
            "sane",
            "socketio",
            "stdnum",
            "string_utils",
            "swiftclient",
            "tackerclient",
            "uinput",
            "vitrageclient",
            "xmp-toolkit",
            "zaqarclient",
        ]:
            return "python-" + name

        if name in [
            "acoustid",
            "alsa",
            "alsaaudio",
            "collada",
            "cpuinfo",
            "enchant",
            "hamcrest",
            "smbc",
            "smi",
            "svg",
            "winrm",
        ]:
            return "py" + name

        if name in ["nose-xcover", "paste-script", "paste-deploy"]:
            return name.replace("-", "")

        if name in ["btchip", "edgegrid", "mmtf"]:
            return name + "-python"

        return PortingdbLoader._pypi_name_mapping.get(name, name)

    @classmethod
    def weblists(cls) -> dict:
        return {
            "fedora": {
                "getter": PortingdbLoader._get_fedora,
                "parser": PortingdbLoader._parse_fedora,
            }
        }

    @classmethod
    def load(cls, name: str, **kwargs) -> Dataset:
        # get the request type (weblist vs. organization)
        from_type = kwargs.pop("from_type")
        if from_type in ["user", "org"]:
            raise Exception(
                "portingdb does not support loading project lists from user/org names."
            )

        # initialize a registry
        ds = Dataset(**kwargs)

        # select the correct weblist loader/parser
        weblists = cls.weblists()
        if name not in weblists:
            raise Exception(
                "Unrecognized portingdb weblist name. Valid "
                "options are: %s" % list(weblists)
            )

        # load the data
        data = weblists[name]["getter"](api=ds.api, **kwargs)

        # parse the data
        weblists[name]["parser"](ds, data)

        return ds

    @staticmethod
    def _get_fedora(api, **kwargs) -> list:
        url = "https://raw.githubusercontent.com/fedora-python/portingdb/master/data/fedora.json"

        status, data = api.request(url, **kwargs)
        if status != 200:
            raise Exception("Error downloading %s; is the url accessible?", url)

        for name, record in data.items():
            record["project"] = PortingdbLoader.get_pypi_name(name)

        # python- is only ~2000 of ~3800
        return [record for name, record in data.items() if record["project"]]

    @staticmethod
    def _parse_fedora(ds: Dataset, data: list) -> None:
        # map data keys to project keywords
        uuids = {"name": lambda p: p.project}

        # create the projects
        ds.projects = [
            PypiProject(uuids_=uuids, **d)
            for d in tqdm(data, desc="         Loading", unit="project", leave=False)
        ]


class OpenSUSEOBSLoader(Loader):

    _pypi_name_mapping = {
        "abseil": "absl-py",
        "antlr3_runtime": "antlr3-python-runtime",
        "dialog": "python2-pythondialog",
        "fastTSNE": "openTSNE",
        "pocketsphinx-python": "pocketsphinx",
    }

    _not_pypi = [
        "aci-integration-module",  # https://github.com/noironetworks/aci-integration-module
        "atomicwrites-doc",
        "avocado-plugins-vt",
        "babel-doc",
        "distutils-extra",  # http://launchpad.net/python-distutils-extra/
        "dukpy-kovidgoyal",  # https://github.com/kovidgoyal/dukpy ; conflicts with dukpy
        "epubmerge",  # https://github.com/JimmXinu/EpubMerge/issues/5
        "espeak",  # https://launchpad.net/python-espeak
        "espressomd",
        "espressopp",  # https://github.com/espressopp/espressopp/issues/66
        "goocanvas",  # https://github.com/GNOME/pygoocanvas
        "gtksourceview",
        "jaraco.base",
        "jedihttp",  # https://github.com/vheon/JediHTTP/issues/17
        "kasa",
        "killswitch",  # http://blog.homac.de
        "liblarch",  # https://live.gnome.org/liblarch
        "linux-procfs",  # http://userweb.kernel.org/python-linux-procfs (also mirrored on github)
        "nose-random",  # https://github.com/xlwings/nose-random/issues/2
        "oci-sdk",
        "onionshare",  # https://github.com/micahflee/onionshare/issues/910 (packaging related)
        "pycxx",  # http://cxx.sourceforge.net/
        "pygments-style-railscasts",  # https://github.com/DrMegahertz/pygments-style-railscasts is 404
        #                               https://github.com/JoeyButler/pygments-style-railscasts might be ok
        "pyGRID",  # that name is now PySyft, but https://github.com/dedalusj/pyGRID is unrelated
        "pyotherside",  # http://thp.io/2011/pyotherside/
        "pysvn",  # http://pysvn.tigris.org/
        "sge-pygame",  # http://stellarengine.nongnu.org
        "testtools-doc",
        "typeshed",  # package deleted from PyPI
        "virtkey",  # https://launchpad.net/virtkey/
        "xcaplib",  # https://github.com/AGProjects/python-xcaplib
        "zope.proxy-doc",
    ]

    @staticmethod
    def get_pypi_name(package_name):
        if package_name.endswith("rpm-macros"):
            return

        if package_name.startswith("python-"):
            name = package_name[7:]
        elif package_name.startswith("python2-") or package_name.startswith("python3-"):
            name = package_name[8:]
        else:
            return

        # multiple packages, per version
        if name[-1].isdigit():
            if name.startswith("tornado"):
                return
            if name.startswith("pytest"):
                return
            if name.startswith("eric"):
                return
            if name.startswith("Django"):
                return
            if name.startswith("prompt_toolkit"):
                return
            if name.startswith("wxpython"):
                return

        if name.lower() in OpenSUSEOBSLoader._not_pypi:
            return

        if name.startswith("xsge-") or name.startswith("xsge_"):
            return

        if name == "azure-sdk":
            return  # meta package

        if name.lower() in [
            "3parclient",
            "axolotl",
            "axolotl-curve25519",
            "debian",
            "djvulibre",
            "iptables",
            "keyczar",
            "levenshtein",
            "markdown-math",
            "mhash",
            "nss",
            "sane",
            "yamldoc",
        ]:
            return "python-" + name

        if name.lower() in [
            "cairo",
            "cddb",
            "gpgme",
            "gtk",
            "opengl",
            "opengl-accelerate",
            "parted",
            "usb",
        ]:
            return "py" + name

        if name.lower() in OpenSUSEOBSLoader._not_pypi:
            return

        return OpenSUSEOBSLoader._pypi_name_mapping.get(name, name)

    @classmethod
    def weblists(cls) -> dict:
        return {
            "devel:languages:python": {
                "getter": OpenSUSEOBSLoader._get_project,
                "parser": OpenSUSEOBSLoader._parse_project,
            },
            "devel:languages:python:pytest": {
                "getter": OpenSUSEOBSLoader._get_project,
                "parser": OpenSUSEOBSLoader._parse_project,
            },
            "devel:languages:python:numeric": {
                "getter": OpenSUSEOBSLoader._get_project,
                "parser": OpenSUSEOBSLoader._parse_project,
            },
            "devel:languages:python:django": {
                "getter": OpenSUSEOBSLoader._get_project,
                "parser": OpenSUSEOBSLoader._parse_project,
            },
            "devel:languages:python:flask": {
                "getter": OpenSUSEOBSLoader._get_project,
                "parser": OpenSUSEOBSLoader._parse_project,
            },
            "devel:languages:python:jupyter": {
                "getter": OpenSUSEOBSLoader._get_project,
                "parser": OpenSUSEOBSLoader._parse_project,
            },
            "devel:languages:python:avocado": {
                "getter": OpenSUSEOBSLoader._get_project,
                "parser": OpenSUSEOBSLoader._parse_project,
            },
            "devel:languages:python:mailman": {
                "getter": OpenSUSEOBSLoader._get_project,
                "parser": OpenSUSEOBSLoader._parse_project,
            },
            "devel:languages:python:certbot": {
                "getter": OpenSUSEOBSLoader._get_project,
                "parser": OpenSUSEOBSLoader._parse_project,
            },
            "devel:languages:python:azure": {
                "getter": OpenSUSEOBSLoader._get_project,
                "parser": OpenSUSEOBSLoader._parse_project,
            },
            "devel:languages:python:aws": {
                "getter": OpenSUSEOBSLoader._get_project,
                "parser": OpenSUSEOBSLoader._parse_project,
            },
            "devel:languages:python:misc": {
                "getter": OpenSUSEOBSLoader._get_project,
                "parser": OpenSUSEOBSLoader._parse_project,
            },
        }

    @classmethod
    def load(cls, name: str, **kwargs) -> Dataset:
        # get the request type (weblist vs. organization)
        from_type = kwargs.pop("from_type")
        if from_type in ["user", "org"]:
            raise Exception(
                "opensuse OBS does not support loading project lists from user/org names."
            )

        # initialize a registry
        ds = Dataset(**kwargs)

        # select the correct weblist loader/parser
        weblists = cls.weblists()
        if name not in weblists:
            raise Exception(
                "Unrecognized opensuse OBS weblist name. Valid "
                "options are: %s" % list(weblists)
            )

        # load the data
        data = weblists[name]["getter"](api=ds.api, project=name, **kwargs)

        # parse the data
        weblists[name]["parser"](ds, data)

        return ds

    @staticmethod
    def _get_project(api, project, **kwargs) -> list:
        url = "https://build.opensuse.org/package?length=2000&project="
        url += project

        status, data = api.request(url, **kwargs)
        if status != 200:
            raise Exception("Error downloading %s; is the url accessible?", url)

        data = data["data"]

        for record in data:
            name = record["name"]
            start_pos = name.find(">") + 1
            end_pos = name.find("<", start_pos)
            name = name[start_pos:end_pos]
            record["project"] = OpenSUSEOBSLoader.get_pypi_name(name)

        return [record for record in data if record["project"]]

    @staticmethod
    def _parse_project(ds: Dataset, data: list) -> None:
        # map data keys to project keywords
        uuids = {"name": lambda p: p.project}

        # create the projects
        ds.projects = [
            PypiProject(uuids_=uuids, **d)
            for d in tqdm(data, desc="         Loading", unit="project", leave=False)
        ]


webloader_map["portingdb"] = PortingdbLoader
registry_map["portingdb"] = registry_map["pypi"]
api_map["portingdb"] = api_map["pypi"]

webloader_map["opensuse"] = OpenSUSEOBSLoader
registry_map["opensuse"] = registry_map["pypi"]
api_map["opensuse"] = api_map["pypi"]


def get_top_packages(kind="top4kmonth"):
    ds = Dataset.load_web(
        name=kind,
        from_type="list",
        registry="pypi",
        cache_dir=R2C_WEB_CACHE,
    )

    for project in ds.projects:
        yield project.get_name()


def get_fedora_packages():
    ds = Dataset.load_web(
        name="fedora",
        from_type="list",
        registry="portingdb",
        cache_dir=R2C_WEB_CACHE,
    )

    names = set([project.get_name() for project in ds.projects])
    return names


def get_opensuse_packages(project):
    ds = Dataset.load_web(
        name=project,
        from_type="list",
        registry="opensuse",
        cache_dir=R2C_WEB_CACHE,
    )

    # Avoid dups like python2-cmd2 and python-cmd2
    names = set([project.get_name() for project in ds.projects])
    return names


def _intersect_stdlib(kind="top4kyear"):
    for name in get_top_packages(kind):
        if name not in _stdlib_all:
            continue
        yield name



import collections
import logging
import re

import textdistance
from logging_helper import setup_logging

from pypidb._compat import PY2, logger_helper, urlsplit

logger = setup_logging()
if logger_helper and not PY2:
    autolog = logger_helper.LoggerHelper(logging.getLogger(__name__), logging.DEBUG)
else:  # pragma: no cover
    autolog = lambda x: x

_similarity_func = textdistance.RatcliffObershelp().distance


def normalize(name):
    return re.sub(r"[-_.]+", "-", name).lower()


@autolog
def _compute_similarity(name, url, algo=_similarity_func, comp_op=min):
    name = normalize(name).replace("-", "")

    p = urlsplit(url)
    path = p.path
    path = path[1:]
    first, _, second = path.rstrip("/").rpartition("/")
    first = first.replace("/", ".")
    first = normalize(first).replace("-", "")
    second = normalize(second).replace("-", "")

    if second:
        if second.endswith(".git"):  # pragma: no cover
            second = second[:-4]
    elif first.endswith(".git"):  # pragma: no cover
        first = first[:-4]

    if first.startswith("?p="):  # pragma: no cover
        first = first[3:]

    logger.debug("_similarity of {} {} to {}".format(first, second, name))

    last = second if second else first

    results = [
        algo(name, second),
        algo(name, first + second),
        algo(name, second + first),
        algo("py" + name, second),
        algo("py" + name, first + second),
        algo(name + "py", second),
        algo(name + "py", first + second),
        algo(name + "python", second),
        algo(name + "python", first + second),
        algo(name, last + "python"),
        algo(name, "python" + last),
        algo(name, last.replace("python", "").replace("--", "-").strip("-")),
        algo(
            name,
            last.replace("python", "").replace("sdk", "").replace("--", "-").strip("-"),
        ),
        algo(name, last.replace("sdk", "").replace("--", "-").strip("-")),
        algo(name, last + "py"),
        algo(name, "py" + last),
        algo(name, "backports-" + last),
        algo("python" + name + "client", last),  # yubico
        algo("python" + name, last + "client"),
        algo(name + "client", last),
        algo(name + "python-client", last),
        algo(name + "client-python", last),
        algo(name, last + "python-client"),
        algo(name, last + "client"),
        algo(name, last + "client-python"),
        0.1 + algo(name, last + "-ctypes"),  # yara
        algo(
            "backports-" + name, last + "ffi"
        ),  # https://github.com/r3m0t/backports.lzma
        0.1 + algo(name, "django-" + last),  # django-vz-wiki
        0.1 + algo("django-" + name, last),
        0.1 + algo(name.replace("django", "drf"), last),  # django-extra-fields
        0.1 + algo(name.replace("drf", "django"), last),
        algo(name.replace("drf", "django-rest-framework"), last),  # drf-jwt
        algo(name, last.replace("drf", "django-rest-framework")),
    ]

    return comp_op(results)


@autolog
def _get_weighted_best(name, urls, cond):
    name_cleaned = name.lower().replace("_", "").replace("-", "")

    counter = collections.Counter(url for url in urls if cond(name_cleaned, url))

    if not counter:
        return

    logger.debug("counter - {}".format(counter))

    most_common = counter.most_common()

    if len(most_common) == 1:  # pragma: no cover
        return most_common[0][0]

    # ignore items of same weight
    if most_common[0][1] == most_common[1][1]:
        return [i[0] for i in most_common if i[1] == most_common[0][1]]

    return most_common[0][0]


@autolog
def _get_most_similar(name, urls):
    data = dict((url, _compute_similarity(name, url)) for url in set(urls))
    logger.debug("computed similarity: {}".format(data))

    nearest_value = None
    nearest_counter = collections.Counter()
    nearest_url = None
    for url, value in data.items():
        nearest_counter[value] += 1
        if nearest_value is None or value < nearest_value:
            logger.debug("nearest: {} {}".format(value, url))
            nearest_value = value
            nearest_url = url
    logger.debug("nearest_counter {}".format(nearest_counter))
    if nearest_counter[nearest_value] != 1:
        # Identical similarity for different values
        logger.debug(
            "identical similarity {} times: {}".format(
                nearest_counter[nearest_value], nearest_value
            )
        )
        return

    return nearest_url


def clean_url(url):
    return url.lower().replace("_", "").replace("-", "").rstrip("/")


def _slash_endswith(name_cleaned, url):
    return clean_url(url).endswith("/" + name_cleaned)


def _endswith(name_cleaned, url):
    return clean_url(url).endswith(name_cleaned)


def _python_and_name_match(name_cleaned, url):
    url = clean_url(url)
    return "python" in url and name_cleaned in url


def _only_name_match(name_cleaned, url):
    return name_cleaned in clean_url(url)


@autolog
def _get_shortest(urls):
    shortest = min(len(url) for url in urls)
    urls = [url for url in urls if len(url) == shortest]
    if len(urls) > 1:
        return

    return urls[0]


@autolog
def _get_prefered(urls):
    orig_urls = urls

    freedesktop_urls = [url for url in urls if "gitlab.freedesktop.org" in url]
    if len(freedesktop_urls) == 1:
        return freedesktop_urls[0]

    freedesktop_urls = [url for url in urls if "cgit.freedesktop.org" in url]
    if len(freedesktop_urls) == 1:
        return freedesktop_urls[0]

    opendev_urls = [url for url in urls if "opendev.org" in url]
    if len(opendev_urls) == 1:
        return opendev_urls[0]

    urls = [url for url in urls if "code.google.com" not in url]
    if len(urls) == 1:
        return urls[0]
    elif not urls:  # pragma: no cover
        urls = orig_urls

    urls = [url for url in urls if "sourceforge.net" not in url]
    if len(urls) == 1:
        return urls[0]
    elif not urls:  # pragma: no cover
        urls = orig_urls

    shortest = _get_shortest(urls)
    if shortest:
        logger.debug("using shortest")
        return shortest

    logger.debug("using urls[0]")
    return urls[0]


@autolog
def get_best_match(name, urls):
    accepted = urls

    best = _get_weighted_best(name, accepted, _slash_endswith)

    if best:
        if isinstance(best, str):
            return best
        accepted = best

    best = _get_weighted_best(name, accepted, _endswith)

    if best:
        if isinstance(best, str):
            return best
        accepted = best

    best = _get_weighted_best(name, accepted, _python_and_name_match)

    if best:
        if isinstance(best, str):
            return best
        accepted = best

    best = _get_weighted_best(name, accepted, _only_name_match)

    if best:
        if isinstance(best, str):
            return best
        accepted = best

    most_common = _get_weighted_best(name, accepted, lambda x, y: True)

    logger.debug("!!!!finding similar: {}".format(accepted))
    most_similar = _get_most_similar(name, set(accepted))

    # Reject multiple
    if not isinstance(most_common, str):
        accepted = most_common
        most_common = None

    if not most_common and not most_similar:
        return _get_prefered(accepted)

    logger.debug("common, similar: {} - {}".format(most_common, most_similar))
    if most_common == most_similar:
        logger.debug("most common and most similar: {}".format(most_common))
        return most_common
    elif most_common:
        logger.debug("most common: {}".format(most_common))
        return most_common
    elif most_similar:
        logger.debug("most similar w/ identical commonality: {}".format(most_similar))
        return most_similar
    else:  # pragma: no cover
        return _get_prefered(accepted)



import requests
from logging_helper import setup_logging
from requests.packages.urllib3.util.timeout import Timeout
from requests.utils import _parse_content_type_header

from https_everywhere.adapter import BlockAdapter, RedirectAdapter, _generate_response
from pypidb._compat import urlsplit

logger = setup_logging()


def get_file_lines(filename):
    with open(filename) as f:
        return set(i.lower() for i in f.read().splitlines() if not i[0] == "#")


class Status500Adapter(BlockAdapter):

    send = BlockAdapter.send_block


class IPBlockAdapter(BlockAdapter, RedirectAdapter):
    def send(self, request, *args, **kwargs):
        url = request.url
        if url.startswith("http"):
            if url[4] == "s":
                start = 8
            else:
                start = 7
            is_num = url[start].isdigit()
            is_IP = False
            logger.debug("is_num = {}; is_IP = {}: {}".format(is_num, is_IP, url))
            if is_num:
                ip_parts = url[start:].split(".", 3)
                logger.debug("ip_parts = {}: {}".format(ip_parts, url))
                if len(ip_parts) == 4:
                    ip_parts[3] = ip_parts[3].split("/", 1)[0]
                    logger.debug("ip_parts = {}: {}".format(ip_parts, url))
                    is_IP = all(part.isdigit() for part in ip_parts)
                elif ip_parts[0].startswith("localhost"):
                    is_IP = True
            logger.debug("is_IP = {}: {}".format(is_IP, url))
            if is_IP:
                return self.send_block(request, *args, **kwargs)

        logger.debug("IPblock of {} skipped".format(request.url))
        try:
            resp = super(IPBlockAdapter, self).send(request, *args, **kwargs)
        except Exception as e:
            resp = self.handle_error(e)
        return resp


class CDNBlockAdapter(BlockAdapter):
    cdn_prefixes = {
        "ajax",
        "assets",
        "cdn",
        "cdnjs",
        "fonts",
        "i",
        "img",
        "images",
        "imgix",
        "maxcdn",
        "media",
        "netdna",
        "static",
        "unpkg",  # unpkg.com
    }
    cdn_domains = {
        "a.trellocdn.com",
        "code.jquery.com",
    }

    def send(self, request, *args, **kwargs):
        url = request.url
        p = urlsplit(url)
        netloc = p.netloc.lower()
        parts = netloc.split(".")
        if parts[0] in self.cdn_prefixes or netloc in self.cdn_domains:
            return self.send_block(request, *args, **kwargs)

        logger.debug("cdn block of {} skipped".format(request.url))
        try:
            resp = super(CDNBlockAdapter, self).send(request, *args, **kwargs)
        except Exception as e:
            resp = self.handle_error(e)
        return resp


class DomainListBlockAdapter(BlockAdapter):
    def __init__(self, *args, **kwargs):
        blocklist = kwargs.pop("blocklist", None)
        super(DomainListBlockAdapter, self).__init__(*args, **kwargs)
        self._blocklist = None
        if blocklist:
            self._blocklist = get_file_lines(blocklist)

    def send(self, request, *args, **kwargs):
        url = request.url
        p = urlsplit(url)
        netloc = p.netloc.lower()
        for pattern in self._blocklist:
            if pattern in netloc:
                return self.send_block(request, *args, **kwargs)

        logger.debug("domain block of {} skipped".format(request.url))
        try:
            resp = super(DomainListBlockAdapter, self).send(request, *args, **kwargs)
        except Exception as e:
            resp = self.handle_error(e)
        return resp


class LoginBlockAdapter(BlockAdapter):

    _login_redirect_patterns = ["login", "signin"]
    _login_site_prefix = [
        "https://accounts.google.com/servicelogin",
        "https://open.login.yahooapis.com/openid/op/auth",
        "https://citeseer.ist.psu.edu/myciteseer/login",
        "http://login.developer.nvidia.com/",  # nvidia-ml-py
        "https://login.newrelic.com/login",  # newrelic
    ]

    def block_redirect(self, from_url, to_url):
        from_url_lower = from_url.lower()
        to_url_lower = to_url.lower()

        for pattern in self._login_redirect_patterns:
            if pattern not in from_url_lower and pattern in to_url_lower:
                return 403

    def send(self, request, *args, **kwargs):
        url_lower = request.url.lower()
        for prefix in self._login_site_prefix:
            if url_lower.startswith(prefix):
                return self.send_block(request, code=403, *args, **kwargs)

        try:
            resp = super(LoginBlockAdapter, self).send(request, *args, **kwargs)
        except Exception as e:
            resp = self.handle_error(e)

        if resp.status_code == 302:
            is_login_redirect = self.block_redirect(request.url, resp.url)

            if is_login_redirect:
                logger.info("blocking redirect to {}".format(resp.url))
                if "location" in resp:
                    del resp.headers["location"]
                resp.status_code = 403

        return resp


class ContentTypeBlockAdapter(BlockAdapter):
    def send(self, request, *args, **kwargs):
        try:
            resp = super(ContentTypeBlockAdapter, self).send(request, *args, **kwargs)
        except Exception as e:
            resp = self.handle_error(e)

        location = resp.headers.get("location")
        if location:
            return resp

        content_type = resp.headers.get("content-type")
        if not content_type:
            return self.send_block(request, code=500, *args, **kwargs)

        content_type, params = _parse_content_type_header(content_type)

        if content_type == "text/plain":
            try:
                resp.text
            except Exception as e:
                logger.error("text/plain which isnt plain text: {}".format(e))
                return self.send_block(request, code=406, *args, **kwargs)

        parts = content_type.split("/")
        if parts[0] not in ["text", "application"]:
            return self.send_block(request, code=406, *args, **kwargs)

        if parts[1] in ["json"]:
            return resp

        if parts[1] in ["javascript", "css"]:
            return self.send_block(request, code=406, *args, **kwargs)

        if parts[0] == "application":
            if "html" not in parts[1] and "xml" not in parts[1]:
                return self.send_block(request, code=406, *args, **kwargs)

        return resp


class HTTPSAdapter(RedirectAdapter):

    _head_timeout = Timeout(connect=10, read=5)

    def __init__(self, *args, **kwargs):
        https_exceptions = kwargs.pop("https_exceptions", [])
        super(HTTPSAdapter, self).__init__(*args, **kwargs)
        self._https_exceptions = https_exceptions

    def ignore_handle_error(self, exc, request=None):
        if not request:
            raise exc

        if request.url.startswith("https"):
            tail = request.url[8:]
            if self.prevent_https(tail):
                logger.info("downgrading to http", request.url, exc)
                new_request = request.copy()
                new_request.url = "http://" + tail
                return self.send(request, timeout=self._head_timeout)

        raise exc

    def block_redirect(self, from_url, to_url):
        pass

    def prevent_https(self, tail):
        if tail[0].isdigit():  # TODO: detect IP
            return True
        for rule in self._https_exceptions:
            if tail.startswith(rule):
                return True
        return False

    def send(self, request, *args, **kwargs):
        if request.url.startswith("https://"):
            tail = request.url[8:]
            if self.prevent_https(tail):
                logger.info("downgraded {} to http".format(request.url))
                request.url = "http://" + tail

        try:
            resp = super(HTTPSAdapter, self).send(request, *args, **kwargs)
        except Exception as e:
            resp = self.handle_error(e, request)
        return resp

    def get_redirect(self, url):
        if not url.startswith("http://"):
            return super(HTTPSAdapter, self).get_redirect(url)

        tail = url[7:]
        if self.prevent_https(tail):
            logger.info("https blocked for {}".format(url))
            return super(HTTPSAdapter, self).get_redirect(url)

        while True:
            current_url = url
            try:
                if current_url.startswith("http://code.google.com"):
                    resp = requests.get(
                        current_url, allow_redirects=False, timeout=self._head_timeout
                    )
                else:
                    resp = requests.head(
                        current_url, allow_redirects=False, timeout=self._head_timeout
                    )
                resp.raise_for_status()
            except Exception as e:
                logger.info("head failed for {}: {!r}".format(current_url, e))
                break
            else:
                logger.debug(
                    "head {} {} {} {} {}".format(
                        current_url, resp.url, resp, resp.headers, resp.content
                    )
                )
                location = resp.headers.get("location")
                if location and location != current_url:
                    code = self.block_redirect(current_url, location)
                    if code is True:
                        return super(HTTPSAdapter, self).get_redirect(url)
                    elif code:
                        return _generate_response(code)

                    code = self.block_redirect(url, location)
                    if code is True:
                        return super(HTTPSAdapter, self).get_redirect(url)
                    elif code:
                        return _generate_response(code)

                    if location.startswith("https:"):
                        tail = location[8:]
                        if self.prevent_https(tail):
                            return super(HTTPSAdapter, self).get_redirect(url)
                        return resp
                    url = location

                break

        tail = url[7:]
        if self.prevent_https(tail):
            logger.info("https blocked for {}".format(url))
            return super(HTTPSAdapter, self).get_redirect(url)
        return "https" + url[4:]



from unittest_expander import expand, foreach

from tests.data import exact, mismatch, missing_repos
from tests.utils import _TestBase, normalize, web_session

expected = {}
expected.update(exact)
expected.update(mismatch)


class TestSpecific(_TestBase):
    def test_pylirc(self):
        self.assertRaisesNoUrls("pylirc")
        # "https://sourceforge.net/projects/pylirc"


@expand
class TestExplicit(_TestBase):

    expected = expected
    names = [
        name
        for name, url in expected.items()
        if "sourceforge.net" in url and "sourceforge.net/projects/turbogear1" not in url
    ]
    expected_failures = []

    @foreach(names)
    def test_package(self, name):
        expected = self.expected[name]
        try:
            url = self.converter.get_vcs(name)
        except Exception:
            if name in self.expected_failures:
                return
            raise

        if name in self.expected_failures:
            return

        self.assertIsNotNone(url)
        self.assertInsensitiveEqual(url, expected)

        normalised_name = normalize(name)

        if normalised_name in missing_repos:
            return
        r = web_session.get(url)
        r.raise_for_status()



from pypidb._exceptions import (
    InvalidPackage,
    InvalidPackageVersion,
    IncompletePackageMetadata,
    PackageWithoutFiles,
    PackageWithoutUrls,
    UnrecognisedStdlibBackport,
)
from tests.utils import _TestBase


class TestVersion(_TestBase):
    def test_rfc6555(self):
        # Current release is 0.0.0
        url = self.converter.get_vcs("rfc6555")
        self.assertInsensitiveEqual(url, "https://github.com/SethMichaelLarson/rfc6555")

    def test_zmq(self):
        with self.assertRaises(InvalidPackageVersion):
            self.converter.get_vcs("zmq")

    def test_skimage(self):
        with self.assertRaises(InvalidPackageVersion):
            self.converter.get_vcs("skimage")

    def test_locust(self):
        with self.assertRaises(InvalidPackageVersion):
            self.converter.get_vcs("locust")


class TestFiles(_TestBase):
    def test_django_myrecaptcha(self):
        # Ignore link to https://bitbucket.org/Kizlum/django-myrecaptcha
        with self.assertRaises(PackageWithoutFiles):
            self.converter.get_vcs("django-myrecaptcha")


class TestExceptions(_TestBase):
    def test_missing_package(self):
        with self.assertRaises(InvalidPackage):
            self.converter.get_vcs("reddit")

    def test_stdlib_version_0_0_0_unittest(self):
        with self.assertRaises(InvalidPackageVersion):
            self.converter.get_vcs("unittest")

    def test_stdlib_Wave(self):
        with self.assertRaises(UnrecognisedStdlibBackport):
            self.converter.get_vcs("wave")

    def test_anaconda(self):
        with self.assertRaises(IncompletePackageMetadata):
            self.converter.get_vcs("anaconda")

    def test_bs4(self):
        with self.assertRaises(IncompletePackageMetadata):
            self.converter.get_vcs("bs4")

    def test_dtopt(self):
        with self.assertRaises(PackageWithoutUrls):
            self.converter.get_vcs("dtopt")
        # 'svn.colorstudy.com' is in urls, however it is removed because of dns failure

    def test_no_urls_pykerberos(self):
        with self.assertRaises(IncompletePackageMetadata):
            self.converter.get_vcs("pykerberos")

    def test_no_urls_ana(self):
        with self.assertRaises(IncompletePackageMetadata):
            self.converter.get_vcs("ana")

    def test_no_urls_augeas(self):
        with self.assertRaises(InvalidPackage):
            self.converter.get_vcs("augeas")

    def test_no_urls_dbf(self):
        with self.assertRaises(PackageWithoutUrls):
            self.converter.get_vcs("dbf")

    def test_no_urls_fedorahosted_suds(self):
        with self.assertRaises(PackageWithoutUrls):
            self.converter.get_vcs("suds")

    def test_no_urls_evtx(self):
        with self.assertRaises(IncompletePackageMetadata):
            self.converter.get_vcs("evtx")

    def test_no_urls_inifile(self):
        with self.assertRaises(IncompletePackageMetadata):
            self.converter.get_vcs("inifile")

    def test_no_urls_kid(self):
        with self.assertRaises(IncompletePackageMetadata):
            self.converter.get_vcs("kid")

    def test_no_urls_kmod(self):
        with self.assertRaises(IncompletePackageMetadata):
            self.converter.get_vcs("kmod")

    def test_no_urls_pygments_ansi_color(self):
        with self.assertRaises(IncompletePackageMetadata):
            self.converter.get_vcs("pygments-ansi-color")

    def test_no_urls_pyjwkest(self):
        with self.assertRaises(IncompletePackageMetadata):
            self.converter.get_vcs("pyjwkest")

    def test_no_urls_axolotl_curve25519(self):
        with self.assertRaises(IncompletePackageMetadata):
            self.converter.get_vcs("python-axolotl-curve25519")

    def test_no_urls_qet_tb_generator(self):
        with self.assertRaises(PackageWithoutUrls):
            self.converter.get_vcs("qet_tb_generator")

    def test_no_urls_selection(self):
        with self.assertRaises(IncompletePackageMetadata):
            self.converter.get_vcs("selection")

    def test_no_urls_random2(self):
        with self.assertRaises(PackageWithoutUrls):
            self.converter.get_vcs("random2")

    def test_no_urls_dialog(self):
        with self.assertRaises(PackageWithoutFiles):
            self.converter.get_vcs("dialog")

    def test_no_urls_termcolor(self):
        with self.assertRaises(PackageWithoutUrls):
            self.converter.get_vcs("termcolor")

    def test_no_urls_strict_rfc3339(self):
        with self.assertRaises(PackageWithoutUrls):
            self.converter.get_vcs("strict-rfc3339")

    def test_no_urls_Durus(self):
        with self.assertRaises(PackageWithoutUrls):
            self.converter.get_vcs("Durus")

    def test_no_urls_sane(self):
        with self.assertRaises(PackageWithoutFiles):
            self.converter.get_vcs("sane")

    def test_no_urls_testflo(self):
        with self.assertRaises(IncompletePackageMetadata):
            self.converter.get_vcs("testflo")

    def test_no_urls_moksha(self):
        with self.assertRaises(IncompletePackageMetadata):
            self.converter.get_vcs("moksha-common")
        # Very slow, test_non_found_43, test_non_found_42, test_non_found_39
        # also: test_non_found_81, test_non_found_65
        # cache dns errors?

    def test_no_urls_fedorahosted(self):
        with self.assertRaises(PackageWithoutUrls):
            self.converter.get_vcs("kitchen")

    def test_no_urls_bencode(self):
        with self.assertRaises(PackageWithoutUrls):
            self.converter.get_vcs("bencode")

    def test_no_urls_demjson(self):
        with self.assertRaises(PackageWithoutUrls):
            self.converter.get_vcs("demjson")

    def test_no_urls_kaa(self):
        with self.assertRaises(PackageWithoutUrls):
            self.converter.get_vcs("kaa-base")

    def test_non_found_pythonpaste_org(self):
        with self.assertRaises(PackageWithoutUrls):
            self.converter.get_vcs("Tempita")

    def test_python_org_typing_backport(self):
        with self.assertRaises(PackageWithoutUrls):
            self.converter.get_vcs("typing")

    def test_python_org_url(self):
        with self.assertRaises(IncompletePackageMetadata):
            self.converter.get_vcs("python-docs-theme")

    def test_ovirt_404(self):
        with self.assertRaises(PackageWithoutUrls):
            self.converter.get_vcs("ovirt-engine-sdk")

    def test_ovirt_v2(self):
        with self.assertRaises(IncompletePackageMetadata):
            self.converter.get_vcs("ovirt-engine-sdk-python")

    def test_no_urls_awsebcli(self):
        with self.assertRaises(PackageWithoutUrls):
            self.converter.get_vcs("awsebcli")

    def test_github_issues_tools(self):
        with self.assertRaises(IncompletePackageMetadata):
            self.converter.get_vcs("github-issues-tools")

    def test_non_found_BitVector(self):
        with self.assertRaises(PackageWithoutUrls):
            self.converter.get_vcs("BitVector")
        # https://engineering.purdue.edu/kak/dist/BitVector-3.4.9.html isnt suitable

    def test_non_found_cerealizer(self):
        with self.assertRaises(PackageWithoutUrls):
            self.converter.get_vcs("cerealizer")
        # requires depth 2 to find https://bitbucket.org/jibalamy/cerealizer

    def test_non_found_pybibframe(self):
        with self.assertRaises(PackageWithoutUrls):
            self.converter.get_vcs("pybibframe")

    def test_not_found_abclient(self):
        with self.assertRaises(PackageWithoutUrls):
            self.converter.get_vcs("abclient")

    def test_logentries(self):
        with self.assertRaises(PackageWithoutUrls):
            self.converter.get_vcs("logentries")

    def test_git_ssh_scheme(self):
        with self.assertRaises(IncompletePackageMetadata):
            self.converter.get_vcs("triforce")
        # ignore 'git+ssh://git@github.com/seveas/git-spindle.git'
        # TODO: add check that it was parsed and excluded

    def test_non_found_mysql(self):
        with self.assertRaises(IncompletePackageMetadata):
            self.converter.get_vcs("mysql-connector-python")

    def test_not_found_whois_1(self):
        with self.assertRaises(PackageWithoutUrls):
            self.converter.get_vcs("pythonwhois")
        # depth 2 will find https://github.com/joepie91/python-whois

    def test_non_found_pyweblib(self):
        with self.assertRaises(PackageWithoutUrls):
            self.converter.get_vcs("pyweblib")

    def test_pytidylib(self):
        with self.assertRaises(PackageWithoutUrls):
            self.converter.get_vcs("pytidylib")

    def test_mailgun(self):
        with self.assertRaises(PackageWithoutUrls):
            self.converter.get_vcs("mailgun")

    def test_amazon_dax_client(self):
        with self.assertRaises(IncompletePackageMetadata):
            self.converter.get_vcs("amazon-dax-client")

    def test_logilab_aspects(self):
        with self.assertRaises(PackageWithoutFiles):
            self.converter.get_vcs("aspects")



from unittest_expander import expand, foreach

from pypidb._rules import rules
from pypidb._similarity import normalize
from tests.data import bad_metadata
from tests.utils import _TestBase


def _collate_preload():
    packages = []
    for rule in rules.values():
        if rule.preload:
            packages += [i for i in rule.preload if normalize(i) not in rules]

    return sorted(set(packages))


_indirect_preload_names = _collate_preload()
_direct_names = [i.name for i in rules.values()]
_rule_names = _direct_names + _indirect_preload_names


@expand
class TestRuleKey(_TestBase):

    names = _direct_names
    ignore_missing = [
        i.name
        for i in rules.values()
        if normalize(i.name) in [normalize(i) for i in bad_metadata]
    ]

    @foreach(names)
    def test_package(self, name):
        assert normalize(name) == name
        self._test_names([name], ignore_not_found=name in self.ignore_missing)


@expand
class TestRulePreload(_TestBase):

    names = _indirect_preload_names
    expected_failures = ["project-config", "readthedocs-org"]

    @foreach(names)
    def test_package(self, name):
        assert normalize(name) == name
        self._test_names([name])



from functools import partial
import os.path

from appdirs import user_cache_dir
from logging_helper import setup_logging

import dns_cache.expiration
from dns_cache import NO_EXPIRY, override_system_resolver
from dns_cache.diskcache import DiskCache
from pypidb import __name__ as app_name

from ._cache import cache_subdir
from ._db import multipackage_repos, reverse_mappings
from ._patch import (
    _get_patch_redirects,
    _get_raw_patch_url,
)
from ._url_extract import (
    get_html_hrefs,
    _url_extract_both,
    _url_extractor_wrapper,
    _url_extractor_wrapper_no_dns,
    _url_no_extract,
)
from ._scm_url_cleaner import _match_hostname
from ._similarity import normalize

logger = setup_logging()

dns_cache.expiration.MIN_TTL = NO_EXPIRY

_azure_exclude = [
    "azure-batch-samples",
    "azure-samples",
    "azureml-containers",
    "azureml-sdk-for-r",
    "azuresearch_jfk_files",
    "botframework-solutions",
    "cla-assistant",
    "machinelearningnotebooks",
    "microsoft/ai",
    "mlops",
    "open-telemetry/opencensus-website",  # azure-core-tracing-opencensus
    "open-telemetry/opentelemetry-python",  # "
]


class NoExpirationDiskCache(DiskCache, dns_cache.expiration.NoExpirationCache):
    pass


dns_cache_dirpath = cache_subdir("dns")
dns_cache = NoExpirationDiskCache(directory=dns_cache_dirpath, min_ttl=NO_EXPIRY)
resolver = override_system_resolver(cache=dns_cache)


def preload_reject_match(name, url):
    name = normalize(name)
    existing_result = reverse_mappings.get(url.lower())
    if existing_result and existing_result != name:
        for base in multipackage_repos:
            if url.startswith(base):
                break
        else:
            logger.debug(
                "url {} rejected as it belongs to {}".format(url, existing_result)
            )
            return True
    return False


class Rule(object):
    def __init__(self, name, preload=None, **kwargs):
        self.name = name
        self.key = normalize(name)
        self.match = self.key
        self.preload = preload
        self.fetch_count = 5
        assert preload is None or isinstance(preload, list)
        self.reject_match_func = None
        self.expect_none = False
        self.ignore_bad_metadata = False
        self.patch = False
        self.ignore_urls = []  # FIXME: does not ignore if url has a path
        self.link_extract = _url_extractor_wrapper
        self._data = kwargs
        for key, value in kwargs.items():
            setattr(self, key, value)

    def __str__(self):
        return "{}/{} = {} {!r} {!r}".format(
            self.key, self.match, self.name, self.preload, self._data
        )

    def url_redirects(self):
        if not self.patch:
            return

        url = _get_raw_patch_url(self.patch)

        import requests

        response = requests.get(url)

        assert response.headers["content-type"].startswith(
            "text/plain"
        ), response.headers["content-type"]
        redirects = _get_patch_redirects(response.content, allow_add_only=True)
        assert redirects
        return redirects

    def reject_match(self, name, url):
        name = self.match
        if self.preload:
            if preload_reject_match(name, url):
                return True
        if self.reject_match_func:
            rv = self.reject_match_func(name, url)
            logger.info(
                "{} rejecting match {} {}".format(self.reject_match_func, name, url)
            )
            return rv

    def reject_url(self, name, url):
        if self.ignore_urls:
            for rule in self.ignore_urls:
                require_no_path = "/" not in rule
                if _match_hostname(url, rule, require_no_path=require_no_path):
                    return True

            if url in self.ignore_urls:
                return True
        return False

    def hash(self):
        return self.key


def xstatic_reject_match(name, url):
    if name.lower().startswith("xstatic") and "xstatic" not in url.lower():
        return True


def reject_docs_match(name, url):
    if not name.lower().startswith("docs") and "docs" in url.lower():
        return True


def reject_names(name, url, names):
    for name in names:
        if name in url.lower():
            return True


def ms_azure_reject_match(name, url):
    if url.endswith("azure/azure-sdk"):
        return True
    return reject_names(name, url, _azure_exclude)


def libvirt_hack(name, url):
    if url.startswith("https://gitlab.com/libvirt/libvirt"):
        return url.replace("/-/blob/master/docs/index.html.in", "-python")


def combine(name, url, funcs):
    for func in funcs:
        if func(name, url):
            return True


class DefaultRule(Rule):
    def __init__(self, name, preload=None, **kwargs):
        key = normalize(name)
        if key.startswith("xstatic-"):
            kwargs["reject_match_func"] = xstatic_reject_match
        elif key.startswith("azure") or key == "onedrivesdk":
            if key != "azure":
                preload = ["azure"]
            if key == "azure-kusto-ingest":
                preload = ["azure", "jupyterhub"]
            if key == "azureml-mlflow":
                preload = ["azure", "mlflow"]
            kwargs["reject_match_func"] = partial(
                combine, funcs=[reject_docs_match, ms_azure_reject_match]
            )
            if key == "azure-cosmosdb-table":
                kwargs["repo_filename"] = "build_packages.py"
            elif key.startswith("azure-storage-"):
                kwargs["repo_filename"] = "tool_build_packages.py"
            else:
                kwargs["repo_filename"] = "README.md"
            kwargs["ignore_urls"] = [
                "readthedocs.org",  # azure-monitor
                "dev.azure.com",
            ]
        elif key.startswith("sphinxcontrib-"):
            kwargs["ignore_urls"] = ["sphinx-doc.org", "sphinx.pocoo.org"]
        elif key == "jupyter-js-widgets-nbextension":
            preload = ["jupyter", "jupyterhub", "ipython"]
            kwargs["ignore_urls"] = ["jupyter.org/about", "ipython.org"]
        elif (
            key.startswith("jupyter-")
            or key.startswith("jupyterlab-")
            or key.startswith("nb")
            or key in ["jupyter", "ipysheet"]
        ):
            preload = ["jupyterhub"]
            kwargs["ignore_urls"] = ["jupyter.org"]
        elif (key.startswith("ipy") and key != "ipython") or key in [
            "widgetsnbextension",
            "traittypes",
            "traitlets",
        ]:
            preload = ["ipython"]
            kwargs["ignore_urls"] = ["ipython.org"]
        elif key.startswith("orange-") or key.startswith("orange3-"):
            preload = ["orange3"]
        elif key.startswith("moz"):
            preload = ["sphinx", "sphinx-rtd-theme", "requests"]
        elif key.startswith("kivy-deps-"):
            kwargs["ignore_urls"] = ["kivy.org"]
        elif key in ("turbogears2",):
            preload = ["bzr"]
        elif key in ("turbogears", "turbomail"):
            pass
        elif key.startswith("tg") or key.startswith("turbo"):
            preload = [
                "turbogears2",
                "turbogears",
                "readthedocs-org",
                "fabric",
                "sphinx-rtd-theme",
            ]
        elif key.startswith("gax-google-") or key.startswith("gapic-"):
            preload = ["google-gax", "virtualenv"]
        elif key.startswith("h2o-pysparkling"):
            # https://github.com/h2oai/sparkling-water/issues/1953
            kwargs["ignore_urls"] = ["docs.h2o.ai"]
            preload = ["sphinx-rtd-theme"]
        elif key.startswith("ujson-"):
            kwargs["match"] = key.replace("ujson", "ultrajson")

        super(DefaultRule, self).__init__(name, preload, **kwargs)


class Rules(dict):
    def from_set(self, other):
        for item in other:
            self[item.key] = item


rules = Rules()
rules.from_set(
    {
        Rule("aexpect", ["avocado-framework"]),
        Rule(
            "amazon-dax-client",
            ["boto3", "botocore"],
            ignore_urls=["docs.aws.amazon.com"],
            expect_none=True,
        ),
        Rule("antlr-python-runtime", ignore_bad_metadata=True),
        Rule("antlr3-python-runtime", ["antlr4-python3-runtime"]),
        Rule("anyvc", patch="https://github.com/RonnyPfannschmidt-Attic/anyvc/pull/1"),
        Rule("archinfo", patch="https://github.com/angr/archinfo/pull/83"),
        Rule("aspects", ignore_urls=["www.logilab.org"], expect_none=True),
        Rule(
            "awsebcli",
            ["pip"],
            ignore_urls=["docs.aws.amazon.com", "aws.amazon.com"],
            link_extract=_url_no_extract,
            expect_none=True,
        ),
        Rule("awshelpers", link_extract=_url_no_extract, expect_none=True),
        Rule("aws-sam-translator", ["aws-sam-cli"]),
        Rule("awsretry", ignore_bad_metadata=True),
        Rule("backports", ignore_bad_metadata=True),
        Rule("barnum", ignore_urls=["wordpress.org"]),
        Rule("basemap", ignore_bad_metadata=True),
        Rule("bindep", ["project-config"], expect_none=True),
        Rule("bitvector", ignore_urls=["sun.com"], expect_none=True),
        Rule("bobo", repo_filename="buildout.cfg"),
        Rule(
            "bugsnag", ignore_urls=["www.bugsnag.com", "bugsnag.com"], expect_none=True
        ),
        Rule("bugzilla", link_extract=get_html_hrefs),
        Rule("camelot", ignore_urls=["www.python-camelot.com"], expect_none=True),
        Rule("cassandra-driver", match="datastax-python-driver"),
        Rule("catkin", ignore_bad_metadata=True),
        Rule("ccxt", ignore_urls=["ccxt.trade"]),
        Rule("cdecimal", ignore_bad_metadata=True),
        Rule("celery-redbeat", ["celery"]),
        Rule("cffi", patch="https://foss.heptapod.net/pypy/cffi/commit/9b98da72.patch"),
        Rule("cfscrape", ignore_urls=["nodejs.org"]),
        Rule("cheesecake", link_extract=_url_no_extract, expect_none=True),
        Rule("cheetah", ignore_urls=["pythonhosted.org"]),
        Rule("clickhouse-cityhash", ["cityhash", "xxh", "metrohash"]),
        Rule(
            "cloudshell-automation-api",
            ignore_urls=["www.qualisystems.com"],
            expect_none=True,
        ),
        Rule("coards", ["sphinx-rtd-theme", "zc-buildout"]),
        Rule("cogapp", ignore_urls=["rubyforge.org"]),
        Rule("collectd", ignore_bad_metadata=True),
        Rule("comet-ml", ignore_urls=["comet.ml", "www.comet.ml"], expect_none=True),
        Rule("compatibility-lib", ignore_urls=["www.docker.com"], expect_none=True),
        Rule("compressed-segmentation", ignore_urls=["www.janelia.org"]),
        Rule("config", ignore_urls=["docs.red-dove.com"]),
        Rule("contextual", ignore_bad_metadata=True),
        Rule("cronwatch", ignore_bad_metadata=True),
        Rule(
            "coveralls-check", patch="https://github.com/cjw296/coveralls-check/pull/2"
        ),
        Rule("ctypeslib", ignore_bad_metadata=True),
        Rule("cvxopt", ["sphinx-rtd-theme"], expect_none=True),
        Rule(
            "darcsver",
            ["babel", "tahoe-lafs"],
            ignore_urls=["peak.telecommunity.com", "trac.edgewall.org"],
            expect_none=True,
        ),
        Rule("dash-table", repo_filename="package.json"),
        Rule("databricks-connect", link_extract=_url_no_extract, expect_none=True),
        Rule("datadog-checks-base", match="integrations-core"),
        Rule("dbf", ["epydoc"], expect_none=True),
        Rule(
            "dbus-python",
            ["pyqt5", "pygobject"],
            ignore_urls=["www.riverbankcomputing.co.uk/software/pyqt/intro"],
            expect_none=True,
        ),
        Rule(
            "dbus-signature-pyparsing",
            patch="https://github.com/stratis-storage/dbus-signature-pyparsing/pull/13",
        ),
        Rule("debian", ignore_bad_metadata=True),
        Rule(
            "dephell-pythons", patch="https://github.com/dephell/dephell_pythons/pull/8"
        ),
        Rule(
            "dephell-shells", patch="https://github.com/dephell/dephell_shells/pull/3"
        ),
        Rule("dai-sgqlc-3-5", ["sgqlc"], link_extract=_url_no_extract, expect_none=True),
        Rule("dateutils", patch="https://github.com/jmcantrell/python-dateutils/pull/8"),
        Rule("divide-and-cover", ignore_urls=["unmaintained.tech"], expect_none=True),
        Rule(
            "django-articles",
            patch="https://github.com/codekoala/django-articles/pull/12",
        ),
        Rule("django-common-helpers", ignore_urls=["tivix.com"]),
        Rule("django-datagrid", ignore_urls=["www.agiliq.com"], expect_none=True),
        Rule(
            "django-events", patch="https://github.com/skyl/django-eventpickrs/pull/1"
        ),
        Rule(
            "django-hijack-admin",
            patch="https://github.com/arteria/django-hijack-admin/commit/e0773ff37.patch",
        ),
        Rule("django-inline-ordering", link_extract=_url_no_extract, expect_none=True),
        Rule("django-lfs"),
        Rule("django-lfstheme", ["django-lfs"], expect_none=True),
        Rule(
            "django-portlets",
            patch="https://github.com/diefenbach/django-portlets/pull/2",
        ),
        Rule(
            "django-settings-toml",
            ["django"],
            link_extract=_url_no_extract,
            expect_none=True,
        ),
        Rule(
            "django-staticmediamgr", ignore_urls=["www.crockford.com"], expect_none=True
        ),
        Rule(
            "django-settings-toml",
            patch="https://github.com/maxking/django-settings-toml/pull/5",
        ),
        Rule("django-toolbelt", ignore_urls=["devcenter.heroku.com"], expect_none=True),
        Rule("djtracker", ignore_urls=["www.f4ntasmic.com", "mark-rogers.net"]),
        Rule("dm-xmlsec-binding", ["lxml"], expect_none=True),
        Rule("dox", ["boot2docker"], expect_none=True),
        Rule(
            "drfdocs",
            patch="https://github.com/manosim/django-rest-framework-docs/pull/186",
        ),
        Rule("elementtidy", ignore_bad_metadata=True),
        Rule(
            "encutils",
            ["cssutils"],
            link_extract=get_html_hrefs,
            match="cssutils",
            expect_none=True,
        ),
        Rule("enum", ["enum34", "roundup"], expect_none=True),
        Rule(
            "epl",
            [
                "pip",
                "pipenv",
                "virtualenv",
                "virtualenvwrapper",
                "pipfile",
                "sphinx-rtd-theme",
            ],
            ignore_urls=["virtualenv.pypa.io"],
            expect_none=True,
        ),
        Rule("etsdevtools", ignore_bad_metadata=True),
        Rule("exifread", ["pip", "setuptools"]),
        Rule("fabric", ignore_urls=["fabfile.org", "paramiko.org", "pyinvoke.org"]),
        Rule("fabric2", ignore_urls=["fabfile.org", "paramiko.org", "pyinvoke.org"]),
        Rule("featureflow", patch="https://github.com/JohnVinyard/featureflow/pull/10"),
        Rule("firkin", ["epydoc"], expect_none=True),
        Rule(
            "flake8-immediate",
            patch="https://github.com/schlamar/flake8-immediate/commit/8121d2eb.patch",
        ),
        Rule(
            "flask-compress",
            patch="https://github.com/colour-science/flask-compress/pull/2",
        ),
        Rule("fluidity-sm", ignore_bad_metadata=True),
        Rule("fs", match="pyfilesystem2"),
        Rule(
            "fsspec",
            ["black", "tox", "pre-commit", "tox-conda"],
            link_extract=_url_no_extract,
        ),
        Rule("galaxy", ignore_urls="galaxyproject.org", ignore_bad_metadata=True),
        Rule("galaxy-lib", ["galaxy"]),
        Rule("geopandas", ["shapely"]),
        Rule("getchanges", patch="https://github.com/TheKevJames/getchanges/pull/43"),
        Rule(
            "git-pylint-commit-hook",
            patch="https://github.com/sebdah/git-pylint-commit-hook/pull/74",
        ),
        Rule("github-release", ignore_urls=["dev.hubspot.com"], expect_none=True),
        Rule("gloo", link_extract=_url_no_extract, expect_none=True),
        Rule("gnupginterface", match="py-gnupg", ignore_bad_metadata=True),
        Rule("google-api-core", ["alabaster"]),
        Rule("google-auth", ignore_bad_metadata=True),
        Rule("google-gax", match="gax-python"),
        Rule("google-oauth", ignore_bad_metadata=True),
        Rule("google", ignore_urls=["breakingcode.wordpress.com"], expect_none=True),
        Rule(
            "googleapis-common-protos",
            patch="https://github.com/googleapis/googleapis/pull/204",
            reject_match_func=lambda name, url: "google/googleapis" in url,
        ),
        Rule("grin", patch="https://github.com/rkern/grin/commit/3cda44a6.patch"),
        Rule("gsutil", ignore_urls=["cloud.google.com/storage/docs/gsutil_install"]),
        Rule("hacking", ["pep8"]),
        Rule("haproxyctl", ignore_bad_metadata=True),
        Rule("hs-dbus-signature", patch="https://github.com/stratis-storage/hs-dbus-signature/pull/31"),
        Rule("hypothesis-pytest", ["hypothesis"], expect_none=True),
        Rule("imagecodecs", link_extract=get_html_hrefs),
        Rule("imreg", link_extract=get_html_hrefs),
        Rule("infi-clickhouse-orm", repo_filename="buildout.cfg"),
        Rule("infi-execute", repo_filename="buildout.cfg"),
        Rule(
            "intel-openmp",
            ignore_urls=["software.intel.com"],
            link_extract=_url_no_extract,
            expect_none=True,
        ),
        Rule(
            "ipympl", ["matplotlib"], ignore_urls=["matplotlib.org"], expect_none=True
        ),
        Rule("ipynb", patch="https://github.com/ipython/ipynb/pull/47"),
        Rule("juicer", repo_filename="setup.py.in"),
        Rule(
            "jupyter-c-kernel",
            patch="https://github.com/brendan-rius/jupyter-c-kernel/pull/49",
        ),
        Rule("jupyter-nbformat", ignore_bad_metadata=True),
        Rule("k5test", patch="https://github.com/pythongssapi/k5test/pull/10"),
        Rule("kaitaistruct", ["construct"]),
        Rule("keras-mxnet", ["keras"]),
        Rule("kitchen", ["hashlib"], expect_none=True),
        Rule("libarchive", match="PyEasyArchive"),  # libarchive
        Rule("librato-metrics", ignore_urls=["dev.librato.com"]),
        Rule("libvirt-python", reject_match_func=libvirt_hack),
        Rule("lightblue", ignore_bad_metadata=True),
        Rule("log4tailer", ignore_bad_metadata=True),
        Rule(
            "logentries",
            ["logentries-lecli"],
            ignore_urls=["logentries.com"],
            expect_none=True,
        ),
        Rule("logging", link_extract=_url_no_extract, expect_none=True),
        Rule("magic", ignore_bad_metadata=True),
        Rule("magma-lang", link_extract=_url_no_extract),
        Rule(
            "mailgun",
            ignore_urls=[
                "documentation.mailgun.com",
                "documentation.mailgun.net",
                "mailgun.net",
            ],
            link_extract=_url_no_extract,
            expect_none=True,
        ),
        Rule("mastodon", ignore_bad_metadata=True),
        Rule("maxminddb", match="maxmind-db-reader-python"),
        Rule(
            "membrete",
            ["django-contact-form"],
            ignore_urls=["bazaar-vcs.org"],
            expect_none=True,
        ),
        Rule("meta", ignore_urls=["semver.org"]),
        Rule("mimeparse", ignore_bad_metadata=True),
        Rule("mongoalchemy", expect_none=True),
        Rule("msgpack-python", fetch_count=1),
        Rule("myghty", ["dogpile-cache"], expect_none=True),
        Rule("neptune-client", ["neptune-contrib"], expect_none=True),
        Rule(
            "newrelic",
            ignore_urls=["newrelic.com", "docs.newrelic.com"],
            link_extract=_url_no_extract,
            expect_none=True,
        ),
        Rule(
            "newrelic-plugin-agent",
            patch="https://github.com/MeetMe/newrelic-plugin-agent/commit/e60e98d7.patch",
        ),
        Rule("nimoy-framework", match="nimoy"),  # spock
        Rule("node-semver", match="python-semver"),
        Rule("nose-fixes", ["nose", "nose2", "pytest"], expect_none=True),
        Rule("nvidia-ml-py", ["docutils"], expect_none=True),
        Rule(
            "nxapi-plumbing", patch="https://github.com/ktbyers/nxapi-plumbing/pull/10"
        ),
        Rule("numarray", ignore_bad_metadata=True),
        Rule("optcomplete", ignore_bad_metadata=True),
        Rule("overtest", link_extract=_url_no_extract, expect_none=True),
        Rule(
            "packit",
            ["setuptools", "cython", "pip", "packaging", "platter", "pbr", "wheel"],
            expect_none=True,
        ),
        Rule("pafy", patch="https://github.com/mps-youtube/pafy/pull/249"),
        Rule(
            "parse-accept-language",
            patch="https://github.com/xelven/parse-accept-language/pull/1",
        ),
        Rule("pep8", ["pycodestyle"]),
        Rule("pdfminer3k", ignore_bad_metadata=True),
        Rule("peak", ignore_bad_metadata=True),
        Rule("peak-rules", ignore_bad_metadata=True),
        Rule("pigpio", link_extract=_url_extract_both),
        Rule("pinax", ["django"], expect_none=True),
        Rule("pip-requ", ["setuptools-gitver"]),
        Rule("plaintable", patch="https://github.com/jonathaneunice/plaintable/pull/1"),
        Rule("printdebug", ["colr"], expect_none=True),
        Rule(
            "proboscis", patch="https://github.com/rackerlabs/python-proboscis/pull/28"
        ),
        Rule("py9p", repo_filename="setup.py.in"),
        Rule("pyaml", match="pretty-yaml"),
        Rule("pyastronomy", link_extract=get_html_hrefs, fetch_count=20),
        Rule("pybrain", ignore_bad_metadata=True),
        Rule("pycddb", ignore_bad_metadata=True),
        Rule("pycotap", link_extract=get_html_hrefs),
        Rule(
            "pycountry-convert",
            patch="https://github.com/jefftune/pycountry-convert/pull/1",
        ),
        Rule("pyficache", ["coverage"], repo_filename="__pkginfo__.py"),
        Rule("pyfim", ignore_urls=["www.borgelt.net"], expect_none=True),
        Rule("pygal", patch="https://github.com/Kozea/pygal/pull/494"),
        Rule("pygam", repo_filename="flit.ini"),
        Rule(
            "pygobject",
            link_extract=get_html_hrefs,
            ignore_urls=["ftp.acc.umu.se", "python-gtk-3-tutorial.readthedocs.io"],
        ),
        Rule(
            "pygtk",
            link_extract=get_html_hrefs,
            ignore_urls=["ftp.acc.umu.se", "python-gtk-3-tutorial.readthedocs.io"],
        ),
        Rule("pyke", ignore_bad_metadata=True),
        Rule("pylirc", ignore_urls=["www.lirc.org"]),
        Rule("pymetar", ["future"]),
        Rule("pymf", ignore_bad_metadata=True),
        Rule("pymilia", ignore_urls=["www.boost.org"], ignore_bad_metadata=True),
        Rule("pymilter", patch="https://github.com/sdgathman/pymilter/pull/36"),
        Rule("pymock", link_extract=get_html_hrefs, ignore_bad_metadata=True),
        Rule("pyqt4", ignore_bad_metadata=True),
        Rule("pyrex", ignore_bad_metadata=True),
        Rule(
            "pyro",
            ["pyro4"],
            ignore_urls=["java.sun.com", "www.corba.org"],
            expect_none=True,
        ),
        Rule("pyrouge", ignore_bad_metadata=True),
        Rule("pysphere", ignore_bad_metadata=True),
        Rule("pysmt", ["z3-solver"]),
        Rule("pystatgrab", ignore_bad_metadata=True),
        Rule("pytest-server-fixtures", ["python-jenkins"]),
        Rule("pytext", link_extract=_url_no_extract),
        Rule(
            "python-igraph",
            ["pycairo", "cairocffi"],
            patch="https://github.com/igraph/python-igraph/commit/9adcf87.patch",
        ),
        Rule("python-mhash", ignore_bad_metadata=True),
        Rule("python-owasp-zap-v2-4", match="zap-api-python"),
        Rule(
            "python-pseudorandom",
            patch="https://github.com/smathot/python-pseudorandom/pull/1",
        ),
        Rule("python-qt5", ignore_bad_metadata=True),
        Rule("python-registry", patch="https://github.com/williballenthin/python-registry/pull/97"),
        Rule("pytidylib", link_extract=_url_no_extract, expect_none=True),
        Rule("py-trello", ["tox"], patch="https://github.com/sarumont/py-trello/pull/309"),
        Rule("pyviz-comms", ["pyviz", "jupyterlab"]),
        Rule("pyvows", patch="https://github.com/heynemann/pyvows/pull/135"),
        Rule("pyz3950", ignore_bad_metadata=True),
        Rule("quickproxy", ["tornado-proxy"], expect_none=True),
        Rule("repoze-sphinx-autointerface", ["docutils"]),
        Rule("repoze-who-plugins-sa", match="repoze-who-sqlalchemy"),
        Rule("requests-opentracing", match="python-requests"),
        Rule(
            "requests-threads",
            patch="https://github.com/requests/requests-threads/commit/5cda6.patch",
        ),
        Rule(
            "respite",
            patch="https://github.com/jgorset/django-respite/commit/6c91a4b.patch",
        ),
        Rule("reviewboard", link_extract=_url_no_extract, expect_none=True),
        Rule("rfc6555", ignore_bad_metadata=True),
        Rule("robinhood-aiokafka", ["aiokafka"], link_extract=_url_no_extract, expect_none=True),
        Rule("rook", ignore_urls=["rookout.com"], expect_none=True),
        Rule("rpy2", patch="https://github.com/rpy2/rpy2/commit/ededbbf.patch"),
        Rule("ruffus", patch="https://github.com/cgat-developers/ruffus/pull/118"),
        Rule("s3cmd", patch="https://github.com/s3tools/s3cmd/commit/8d314b01.patch"),
        Rule("s3pip", ["django-tastypie"], ignore_urls=["jlafon.io"], expect_none=True),
        Rule("salt", ignore_urls=["saltstack.org", "repo.saltstack.com"]),
        Rule(
            "scikits-sparse",
            match="scikit-sparse",
            patch="https://github.com/scikit-sparse/scikit-sparse/commit/66b3a518.patch",
        ),
        Rule(
            "scitools",
            patch="https://github.com/hplgit/scitools/pull/45",
            ignore_bad_metadata=True,
        ),
        Rule("schevodurus", ignore_bad_metadata=True),
        Rule("scm-git", ["sampleproject"], expect_none=True),
        Rule("seqdiag", ["nwdiag"]),
        Rule("setuptools-cython", ignore_urls=["mail.python.org"], expect_none=True),
        Rule("shiboken2", ["pyside2"], expect_none=True),
        Rule("shippo", ignore_urls=["goshippo.com"], expect_none=True),
        Rule("should-dsl", patch="https://github.com/nsi-iff/should-dsl/pull/37"),
        Rule("silver-platter", ["dulwich"]),
        Rule("simplegeneric", link_extract=_url_no_extract),
        Rule("simpletal", ignore_bad_metadata=True),
        Rule(
            "singer-python",
            ignore_urls=["Singer.io/", "singer.io", "www.singer.io"],
            expect_none=True,
        ),
        Rule("skl2onnx", ["scikit-learn"]),
        Rule("spark-sklearn", repo_filename="build.sbt"),
        Rule("sphinxcontrib-napoleon", ["numpy"]),
        Rule("sphinxcontrib-youtube", ignore_bad_metadata=True),
        Rule(
            "sshmount-netrc",
            ignore_urls=["github.com"],
            ignore_bad_metadata=True,
            reject_match_func=lambda name, url: url
            == "https://github.com/tjaartvdwalt/tclt",
        ),
        Rule("sst", ["pip", "pip-tools"], match="selenium-simple-test"),
        Rule("stacks", ignore_urls=["jimmyg.org"], expect_none=True),
        Rule("strictyaml", ["pykwalify"]),
        Rule(
            "stsci-distutils", ["pyfits", "drizzlepac"], ignore_urls=["www.stsci.edu"]
        ),
        Rule(
            "stscipython",
            ["drizzlepac"],
            link_extract=_url_no_extract,
            expect_none=True,
            ignore_urls=["www.stsci.edu"],
        ),
        Rule("subgrab", patch="https://github.com/RafayGhafoor/Subscene-Subtitle-Grabber/pull/7"),
        Rule("supercaptcha", ["django"], expect_none=True),
        Rule("swapper", ["vera"]),
        Rule("sysv-ipc", ["posix-ipc"]),
        Rule("tableauhyperapi", ["tableauserverclient"], expect_none=True),
        Rule("telepathy", ["insights"]),
        Rule("tensorboard", repo_filename="package.json"),
        Rule(
            "tgfastdata",
            ["turbogears2", "toscawidgets", "virtualenv"],
            expect_none=True,
        ),
        Rule("tifffile", link_extract=get_html_hrefs),
        Rule("timer2", patch="https://github.com/ask/timer2/pull/3"),
        Rule("toastcord", ["hikari"], expect_none=True),
        Rule(
            "toscawidgets",
            ["turbogears2", "tw2-core", "pyramid", "genshi"],
            expect_none=True,
        ),
        Rule("travis", link_extract=_url_no_extract, expect_none=True),
        Rule("triforce", ["pipsi", "git-spindle"], expect_none=True),
        Rule("tlslite", patch="https://github.com/trevp/tlslite/pull/123"),
        Rule("turbocheetah", ["turbogears2", "fabric", "sphinx-rtd-theme"]),
        Rule("turbogears", ["turbogears2", "pyramid", "genshi"]),
        Rule("tw-dynforms", ["tw2-core"], expect_none=True),
        Rule(
            "tw-forms", ["tw2-core"], ignore_urls=["toscawidgets.org"], expect_none=True
        ),
        Rule("typing", expect_none=True),
        Rule("typogrify-engineer", preload=["typogrify", "engineer"], expect_none=True),
        Rule("ujson-bedframe", match="ultrajson-esnme"),
        Rule("ujson", match="ultrajson"),
        Rule("uuid", link_extract=get_html_hrefs),
        Rule("uwsgitop", patch="https://github.com/xrmx/uwsgitop/pull/60"),
        Rule("uwsgitop", patch="https://github.com/xrmx/uwsgitop/pull/60"),
        Rule("viper", ignore_urls=["fenicsproject.org"], expect_none=True),
        Rule("waiting", ["sphinx-rtd-theme", "flux"]),
        Rule("watson-machine-learning-client", ["alabaster"], expect_none=True),
        Rule("webapp", ignore_bad_metadata=True),
        Rule("webunit", ["docutils"], expect_none=True),
        Rule("wmi", ["pywin32"], expect_none=True),
        Rule("wordaxe", ignore_bad_metadata=True),
        Rule("xpyb", ignore_bad_metadata=True),
        Rule("xml2rfc", ["weasyprint"]),
        Rule("xmltramp", ignore_bad_metadata=True),
        Rule(
            "yubico",
            ["fido2", "yubihsm"],
            ignore_urls=["www.yubico.com", "api.yubico.com"],
            link_extract=_url_no_extract,
        ),
        Rule("yubihsm", ignore_urls=["developers.yubico.com"]),
        Rule("z3c-testsetup", ["zope-app-testing", "zope-testing"]),
        Rule("zabbix-api", ignore_urls=["www.zabbix.com"]),
        Rule("zendesk", patch="https://github.com/maxgutman/zendesk/pull/12"),
        Rule("zsi", ignore_bad_metadata=True),
        Rule(
            "zuul-sphinx",
            ["zuul", "alabaster"],
            ignore_urls=["zuul-ci.org"],
            expect_none=True,
        ),
        Rule("zvmcloudconnector", ["zhmcclient"]),
    }
)

_email_matches = {
    "avocado-devel@redhat.com": "avocado-framework",
    "cgohlke@uci.edu": "cgohlke",
    "colm.oconnor.github@gmail.com": ["hitchdev", "crdoconnor"],
    "developers@pyviz.org": "holoviz/{namespace}_{suffix}",
    "dthomas@osrfoundation.org": ["ros", "ros-infrastructure"],
    "darcy@PyGreSQL.org": "pygresql",
    "efried@us.ibm.com": "powervm",
    "enthought-dev@enthought.com": "srossross",
    "f4nt@f4ntasmic.com": "f4nt",
    "galaxy-dev@lists.galaxyproject.org": "galaxyproject",
    "henning@jacobs1.de": "hjacobs",
    "ian@datagram.co": "boot2docker",
    "ipython-dev@scipy.org": ("ipython", "jupyter-widgets"),
    "jeremy@jeremysanders.net": "veusz",
    "joan@abyz.me.uk": "joan2937",
    "jupyter@googlegroups.com": "jupyter",
    "peak@eby-sarna.com": "PEAK-Legacy",
    "phi@funaba.org": "https://www.funaba.org/code",  # calendar
    "philip@semanchuk.com": "osvenskan",  # sysv_ipc/posix_ipc
    "plaidml-dev@googlegroups.com": "plaidml/plaidml",
    "pygame@pygame.org": "pygame",
    "pypi@tjaart.co.za": "tjaartvdwalt",
    "richard@tartarus.org": "snowballstem",
    "shlomi@smore.com": "smore-inc",
    "smithnick@google.com": "google/{suffix}",  # https://github.com/google/pasta/pull/81
    "support@maxmind.com": "maxmind/{name}-api-python",
    "support@sigopt.com": "sigopt/{name}-python",
    "sympy@googlegroups.com": "sympy",
    "thatch45@gmail.com": "saltstack",
    "tomaz+pypi@tomaz.me": "Yubico/python-{name}",
    "vinay_sajip@red-dove.com": "https://bitbucket.org/vinay.sajip/{name}",
    "vmalloc@gmail.com": "vmalloc",
    "yans@yancomm.net": "zardus",
    "yubico@yubico.com": "Yubico/python-{name}",
}

_namespace_matches = {
    "aliyun": "aliyun/aliyun-openapi-python-sdk",
    "antlr3": "antlr/{namespace}",
    "cognite": "cognitedata",
    "dash": "plotly",
    "dephell": "dephell/dephell_{suffix}",
    "grpcio": (
        "GoogleCloudPlatform/grpc-{suffix}-python",
        "{suffix}-contrib/python-grpc",
        "grpc/grpc",  # should add check for sub-path match of {suffix}(tools, testing, etc)
    ),
    "jupyter": (
        "jupyter",
        "jupyter/jupyter_{suffix}",
        "jupyter/{suffix}",
        "takluyver/jupyter_{suffix}",
        "jdfreder/jupyter-{suffix}",  # jupyter-pip
    ),
    "ipython": "ipython",
    "ipy": ("ipython", "jupyter-widgets"),
    "j1m": "jimfulton/{suffix}",
    "jupyterlab": "jupyterlab",
    "matrix": "matrix-org",  # https://github.com/matrix-org/matrix-synapse-ldap3
    "neptune": "neptune-ai",
    "orange": "biolab",
    "orange3": "biolab",
    "repoze.who.plugins.sa": "repoze/repoze.who-sqlalchemy",
    "repoze": (
        "repoze",
        "repoze/{namespace}.{suffix_parts[0]}-{suffix_parts[2]}",  # repoze.what.plugins.sql -> repoze.what-sql
    ),
    "sphinxcontrib": ("sphinx-doc", "sphinx-contrib"),
    "stsci": "spacetelescope",
    "tg": ("TurboGears", "TurboGears/tg2{suffix}"),
    "tgext": "TurboGears",
    "turbogears2": "Turbogears/tg2/",  # trailing '/' used to only match 'turbogears2', no suffixes
    "xstatic": (
        "openstack",
        "takluyver",
        "xstatic-py",
        "r1chardj0n3s",
        "python-xstatic/{suffix}",
        "dmsimard/{name}-distgit",
        "dmsimard/{suffix}-distgit",
        "dmsimard/python-{name}-distgit",
    ),
    "zc": "zopefoundation",
    "z3c": "zopefoundation",
}


def _find_named_repo(name, emails=None):
    if name.lower() == "xstatic-jquery-ui":
        name = "xstatic-jquery_ui"

    pos_dot = name.find(".")
    pos_dash = name.find("-")
    pos_underscore = name.find("_")
    pos = len(name)
    if pos_dot != -1:
        pos = pos_dot
    if pos_dash != -1 and pos_dash < pos:
        pos = pos_dash
    if pos_underscore != -1 and pos_underscore < pos:
        pos = pos_underscore

    namespace = name[:pos].lower()
    suffix = name[pos + 1 :].lower()

    rules = []
    rule = _namespace_matches.get(name, None)
    if rule:
        if isinstance(rule, str):
            rules.append(rule)
        else:
            rules = list(rule)

    ns_rules = _namespace_matches.get(namespace, [])
    if isinstance(ns_rules, str):
        rules.append(ns_rules)
    else:
        rules += list(ns_rules)

    for email in emails or []:
        address, domain = email.split("@", 1)
        logger.info("email split: {} @ {}".format(address, domain))
        if domain == "lists.sourceforge.net":
            sf_project_name, list_type = address.rsplit("-", 1)
            if list_type in ["bugs", "discuss"]:
                return "https://sourceforge.net/projects/{}".format(sf_project_name)
        elif domain == "lists.freedesktop.org":
            return "https://gitlab.freedesktop.org/{}/{}".format(address, name)

        email_rules = _email_matches.get(email, [])
        if isinstance(email_rules, str):
            email_rules = [email_rules]
        for rule in email_rules:
            if rule not in rules:
                rules.append(rule)

    if not rules:
        return

    logger.info("found repo name rules for {} {}: {!r}".format(name, emails, rules))

    for slug in rules:
        from pypidb._github import check_repo

        if "/" not in slug:
            slug = "{}/{}".format(slug, name)
        elif slug.endswith("/"):
            slug = slug[:-1]
            if namespace != name.lower():
                continue
        else:
            suffix_normalised = normalize(suffix)
            suffix_parts = suffix_normalised.split("-")
            slug = slug.format(
                name=name,
                namespace=namespace,
                suffix=suffix,
                suffix_normalised=suffix_normalised,
                suffix_parts=suffix_parts,
            )

        logger.info("checking {}".format(slug))
        if slug.startswith("https://"):
            return slug
        if check_repo(slug):
            return "https://github.com/{}".format(slug)



from ._compat import urljoin


def get_html_hrefs(content, source_url, only_unique=False, ignore="github.io"):
    from lxml import html

    urls = []
    tree = html.fromstring(content.encode("utf-8"))
    for e in tree.xpath("//*[@href]"):
        link = e.attrib.get("href")
        if len(link) < 2:
            continue
        if link[0] == "/" or link[0] == ".":
            if not source_url:  # pragma: no cover
                continue
            if ignore and ignore in source_url:  # pragma: no cover
                continue
            link = urljoin(source_url, link)
        elif link[0] == "#":  # pragma: no cover
            continue
        elif link.startswith("javascript:"):  # pragma: no cover
            continue
        elif ignore and ignore in link:  # pragma: no cover
            continue

        if only_unique and link in urls:  # pragma: no cover
            continue

        link = urljoin(source_url, link)

        urls.append(link)

    return urls



import unittest

import requests
from cachecontrol import CacheControl, CacheController
from cachecontrol.cache import DictCache
from cachecontrol.heuristics import OneDayCache
from pytest_blockage import block_http, httplib
from requests.packages.urllib3.util.timeout import Timeout

from pypidb._adapters import HTTPSAdapter, IPBlockAdapter
from pypidb._db import mappings
from pypidb._pypi import Converter, InvalidPackage
from tests.utils import _TestBase

http_blocked = hasattr(httplib.HTTPConnection.__init__, "blockage")


# TODO: disable if pytest-blockage enabled globally
# add test that MockHttpCall is raised
class TestPyPiJsonCache(_TestBase):
    def test_miss(self):
        if http_blocked:
            return

        converter = Converter(website_timeout=5)

        with self.assertRaises(InvalidPackage):
            converter.get_vcs("doesntexist")

        try:
            with self.assertRaises(InvalidPackage):
                block_http([])

                converter = Converter(website_timeout=5)
                converter.get_vcs("doesntexist")
        finally:
            httplib.HTTPConnection.__init__ = httplib.HTTPConnection.old

    def test_hit(self):
        if http_blocked:
            return

        converter = Converter(website_timeout=5)
        url = converter.get_vcs("simplejson")
        self.assertEqual(url, "https://github.com/simplejson/simplejson")

        del mappings["simplejson"]

        try:
            block_http([])
            converter = Converter(website_timeout=5)
            url = converter.get_vcs("simplejson")
            self.assertEqual(url, "https://github.com/simplejson/simplejson")
        finally:
            httplib.HTTPConnection.__init__ = httplib.HTTPConnection.old

        try:
            del mappings["simplejson"]
        except KeyError:
            pass

    def test_redirect(self):
        if http_blocked:
            return

        # see also https://github.com/ionrock/cachecontrol/issues/214
        converter = Converter(website_timeout=5)
        url = converter.get_vcs("setuptools_scm")
        self.assertEqual(url, "https://github.com/pypa/setuptools_scm")

        del mappings["setuptools-scm"]

        try:
            block_http([])
            converter = Converter(website_timeout=5)
            url = converter.get_vcs("setuptools_scm")
            self.assertEqual(url, "https://github.com/pypa/setuptools_scm")
        finally:
            httplib.HTTPConnection.__init__ = httplib.HTTPConnection.old

        try:
            del mappings["setuptools-scm"]
        except KeyError:
            pass


class TestFakeRedirect(unittest.TestCase):
    def test_redirect_https(self):
        url = "http://httpbin.org/get?foo"
        session = requests.Session()
        session.mount(url, HTTPSAdapter())
        r = session.get(url)
        self.assertEqual(r.url, url.replace("http://", "https://"))
        self.assertNotIn("location", r.headers)

    def test_redirect_https_south(self):
        url = "http://south.aeracode.org/"
        session = requests.Session()
        session.mount(url, HTTPSAdapter())
        r = session.get(url)
        self.assertEqual(r.url, url.replace("http://", "https://"))
        self.assertNotIn("location", r.headers)
        self.assertIn("bitbucket.org", r.text)

    def test_redirect_https_gustavonarea(self):
        url = "http://code.gustavonarea.net/repoze.who.plugins.sa/"
        session = requests.Session()
        session.mount(url, HTTPSAdapter())
        r = session.get(url)
        self.assertEqual(r.url, url.replace("http://", "https://"))
        self.assertNotIn("location", r.headers)
        self.assertIn("github.com", r.text)

    def _test_no_redirect_https_IP(self):
        url = "http://127.0.0.1/repoze.who.plugins.sa/"
        session = requests.Session()
        session.mount(url, HTTPSAdapter())
        r = session.get(url)
        self.assertEqual(r.url, url)
        self.assertNotEqual(r.url, url.replace("http://", "https://"))

    def test_block_ip(self):
        url = "http://127.0.0.1/repoze.who.plugins.sa/"
        session = requests.Session()
        session.mount("http://", IPBlockAdapter())
        r = session.get(url)
        self.assertEqual(r.status_code, 500)
        self.assertEqual(r.url, url)

    def _test_timeout_moksha_requests_session(self):
        url = "https://mokshaproject.net"
        session = requests.Session()
        r = session.get(url, timeout=Timeout(5))
        self.assertEqual(r.status_code, 200)
        self.assertEqual(r.url, url)

    def _test_timeout_moksha_requests_get(self):
        url = "https://mokshaproject.net"
        r = requests.get(url, timeout=Timeout(5))
        self.assertEqual(r.status_code, 200)
        self.assertEqual(r.url, url)

    def _test_timeout_moksha(self):
        # mokshaproject.net is a parked domain
        url = "https://mokshaproject.net"
        from pypidb._cache import get_file_cache_session

        s = get_file_cache_session("web")
        with self.assertRaises(requests.exceptions.ConnectTimeout):
            s.get(url, timeout=Timeout(5))

    def _test_fetch_ply_requests_session(self):
        url = "https://www.dabeaz.com/ply/"

        s = requests.Session()
        r = s.get(url, timeout=Timeout(5))
        self.assertEqual(r.status_code, 200)
        self.assertEqual(r.url, url)

    def _test_fetch_ply_requests_get(self):
        url = "https://www.dabeaz.com/ply/"

        r = requests.get(url, timeout=Timeout(5))
        self.assertEqual(r.status_code, 200)
        self.assertEqual(r.url, url)

    def test_fetch_ply_https(self):
        from pypidb._cache import get_file_cache_session

        url = "https://www.dabeaz.com/ply/"

        s = get_file_cache_session("web")
        r = s.get(url, timeout=Timeout(5))
        self.assertEqual(r.status_code, 200)
        self.assertNotEqual(r.url, url)
        self.assertEqual(r.url, url.replace("https://", "http://"))

    def test_head_requests_get(self):
        url = "http://gpodder.org/"

        r = requests.head(url, timeout=Timeout(15))
        self.assertEqual(r.status_code, 302)
        self.assertEqual(r.url, url)
        self.assertEqual(r.headers["location"], "https://gpodder.github.io")
        self.assertFalse(r.content)

    def _test_head_cachecontrol(self):
        # cache polluted by HEAD
        url = "http://gpodder.org/"

        class MoreCodesCacheController(CacheController):
            def __init__(self, *args, **kwargs):
                kwargs["status_codes"] = (200, 203, 300, 301, 302, 404)
                super(MoreCodesCacheController, self).__init__(*args, **kwargs)

        cache = DictCache()
        s = CacheControl(
            requests.Session(),
            cache=cache,
            heuristic=OneDayCache(),
            cacheable_methods=["GET", "HEAD"],
            controller_class=MoreCodesCacheController,
        )
        r = s.head(url, timeout=Timeout(15))
        self.assertEqual(r.status_code, 302)
        self.assertEqual(r.url, url)
        self.assertEqual(r.headers["location"], "https://gpodder.github.io")
        self.assertFalse(r.content)
        self.assertFalse(r.from_cache)

        r = s.get(url, timeout=Timeout(15), allow_redirects=False)
        self.assertEqual(r.status_code, 302)
        self.assertEqual(r.url, url)
        self.assertEqual(r.headers["location"], "https://gpodder.github.io")
        self.assertTrue(r.content)
        self.assertIn(b"302 Found", r.content)
        self.assertFalse(r.from_cache)

        # Re-fetch the redirect from the cache
        r = s.get(url, timeout=Timeout(15), allow_redirects=False)
        self.assertEqual(r.status_code, 302)
        self.assertEqual(r.url, url)
        self.assertEqual(r.headers["location"], "https://gpodder.github.io")
        self.assertTrue(r.content)
        self.assertIn(b"302 Found", r.content)

        self.assertTrue(r.from_cache)

        # Now allow redirects

        r = s.get(url, timeout=Timeout(15))
        self.assertEqual(r.status_code, 200)
        self.assertNotEqual(r.url, url)
        self.assertTrue(r.content)
        self.assertFalse(r.from_cache)

        # Re-fetch the redirect from the cache
        r = s.get(url, timeout=Timeout(15), allow_redirects=False)
        self.assertEqual(r.status_code, 302)
        self.assertEqual(r.url, url)
        self.assertEqual(r.headers["location"], "https://gpodder.github.io")
        self.assertTrue(r.content)
        self.assertIn(b"302 Found", r.content)

        self.assertTrue(r.from_cache)

        # Re-fetch the head from the cache
        r = s.head(url, timeout=Timeout(15))
        self.assertEqual(r.status_code, 302)
        self.assertEqual(r.url, url)
        self.assertEqual(r.headers["location"], "https://gpodder.github.io")
        self.assertTrue(r.content)
        self.assertIn(b"302 Found", r.content)
        self.assertTrue(r.from_cache)

    def _test_head_302_bypasses_cache(self):
        # HEAD 302 is completed before the caching
        from pypidb._cache import get_file_cache_session

        url = "http://gpodder.org/"

        s = get_file_cache_session("web")
        r = s.head(url, timeout=Timeout(15))
        self.assertEqual(r.status_code, 302)
        self.assertEqual(r.url, url)
        self.assertEqual(r.headers["location"], "https://gpodder.github.io")
        self.assertFalse(r.content)
        self.assertFalse(hasattr(r, "from_cache"))

        r = s.head(url, timeout=Timeout(15))
        self.assertFalse(hasattr(r, "from_cache"))

    def test_vlc(self):
        from pypidb._cache import get_file_cache_session

        url = "https://wiki.videolan.org/PythonBinding"

        s = get_file_cache_session("web")
        r = s.get(url, timeout=Timeout(15))
        self.assertEqual(r.status_code, 200)
        self.assertEqual(r.url, url)
        self.assertTrue(r.content)
        self.assertIn(b'<meta name="generator" content="MediaWiki', r.content)
        self.assertIn(b"https://git.videolan.org/", r.content)



import os
import os.path
import socket

import requests
from appdirs import user_cache_dir
from cachecontrol import CacheControlAdapter, CacheController
from cachecontrol.caches.file_cache import FileCache
from cachecontrol.heuristics import ExpiresAfter
from logging_helper import setup_logging
from requests.packages.urllib3.util.retry import Retry
from requests.packages.urllib3.util.timeout import Timeout

from pypidb import __name__ as app_name
from ._adapters import (
    CDNBlockAdapter,
    ContentTypeBlockAdapter,
    DomainListBlockAdapter,
    HTTPSAdapter,
    IPBlockAdapter,
    LoginBlockAdapter,
    Status500Adapter,
)
from ._compat import urlsplit

try:
    from future.standard_library import install_aliases

    install_aliases()
except ImportError:  # pragma: no cover
    pass

_block_request_adapter = Status500Adapter()
logger = setup_logging()
_CI = os.getenv("CI", "")

MAX_REDIRECTS = 10
retries = 3
backoff_factor = 0.3
status_forcelist = (500, 502, 504)

retry = Retry(
    total=retries,
    read=retries,
    connect=retries,
    backoff_factor=backoff_factor,
    status_forcelist=status_forcelist,
)


def cache_subdir(name):
    dirpath = os.path.join(user_cache_dir(app_name), name)
    return dirpath


def get_timeout(url):
    if "wiki.ros.org" in url or "abyz.me.uk" in url:
        return Timeout(connect=15, read=20, total=45)
    if "svn.eby-sarna.com" in url or "gitorious.org" in url:  # pragma: no cover
        return Timeout(connect=30, read=30, total=60)
    if "w3.org" in url or "pygal.org" in url:
        return Timeout(connect=30, read=30, total=45)
    if not _CI and "galaxyproject.org" in url:
        return Timeout(connect=100, read=100, total=200)

    return Timeout(connect=15, read=11, total=40)


def _check_url_domain(url):
    if "tracpub.yaco.es" in url:  # django-oot
        return True
    if "hg.hardcoded.net" in url:  # sgmllib3k
        return True
    if "github.org" in url:  # tkreadonly
        return True

    try:
        p = urlsplit(url)
    except Exception:
        return False
    if not p.netloc:
        return False
    try:
        socket.gethostbyname(p.netloc)
    except Exception as e:
        logger.info("domain {} lookup error: {}".format(url, e))
        return False
    return True


class IgnoreVaryExpiresAfter(ExpiresAfter):
    def update_headers(self, response):
        headers = super(IgnoreVaryExpiresAfter, self).update_headers(response)
        if "Vary" in response.headers:
            headers["Vary"] = ""
        return headers


class MoreCodesCacheController(CacheController):
    def __init__(
        self, cache=None, cache_etags=True, serializer=None, status_codes=None
    ):
        status_codes = status_codes or (200, 203, 300, 301, 302, 401, 404)
        super(MoreCodesCacheController, self).__init__(
            cache, cache_etags, serializer, status_codes
        )


class ForceTimeoutHTTPAdapter(
    CDNBlockAdapter,
    DomainListBlockAdapter,
    IPBlockAdapter,
    LoginBlockAdapter,
    ContentTypeBlockAdapter,
    HTTPSAdapter,
    CacheControlAdapter,
):
    def __init__(self, *args, **kw):
        timeout = kw.pop("timeout", None)
        super(ForceTimeoutHTTPAdapter, self).__init__(*args, **kw)
        self.timeout = timeout

    def send(self, request, cacheable_methods=None, timeout=None, **kw):
        if self.timeout and not timeout:  # pragma: no cover
            timeout = self.timeout
        return super(ForceTimeoutHTTPAdapter, self).send(request, timeout=timeout, **kw)


class NoVerifyHTTPAdapter(requests.adapters.HTTPAdapter):
    """An HTTP adapter that ignores TLS validation errors. Use at own risk."""

    def cert_verify(self, conn, url, verify, cert):
        super(NoVerifyHTTPAdapter, self).cert_verify(
            conn=conn, url=url, verify=False, cert=cert
        )


def CacheControl(
    sess,
    cache,
    cache_etags=True,
    serializer=None,
    heuristic=None,
    controller_class=None,
    adapter_class=None,
    cacheable_methods=None,
    **kw
):
    assert cache
    adapter_class = adapter_class or CacheControlAdapter
    adapter = adapter_class(
        cache,
        cache_etags=cache_etags,
        serializer=serializer,
        heuristic=heuristic,
        controller_class=controller_class,
        cacheable_methods=cacheable_methods,
        **kw
    )
    sess.mount("http://", adapter)
    sess.mount("https://", adapter)

    sess.mount("ftp://", _block_request_adapter)
    sess.mount("file://", _block_request_adapter)
    sess.mount("irc://", _block_request_adapter)
    sess.mount("wss://", _block_request_adapter)

    sess.mount("http://readme.md/", _block_request_adapter)
    sess.mount("http://license.md/", _block_request_adapter)
    sess.mount("http://contributing.md/", _block_request_adapter)
    sess.mount("http://manifest.in/", _block_request_adapter)
    sess.mount("http://setup.py/", _block_request_adapter)

    sess.mount("http://p01.notifa.info", _block_request_adapter)
    sess.mount("https://p01.notifa.info", _block_request_adapter)
    sess.mount("http://browsehappy.com", _block_request_adapter)
    sess.mount("https://browsehappy.com", _block_request_adapter)
    sess.mount("http://validator.w3.org/check", _block_request_adapter)
    sess.mount("https://validator.w3.org/check", _block_request_adapter)
    sess.mount(
        "http://www.amazon.com/exec/obidos/external-search", _block_request_adapter
    )  # simplegeneric https://github.com/lipoja/URLExtract/issues/62
    sess.mount("https://www.googletagmanager.com/", _block_request_adapter)
    sess.mount("www.google-analytics.com", _block_request_adapter)
    sess.mount("https://www.youtube.com", _block_request_adapter)
    sess.mount("https://youtu.be", _block_request_adapter)

    sess.mount("http://pypistats.org", _block_request_adapter)
    sess.mount("https://pypistats.org", _block_request_adapter)
    sess.mount("http://docs.python.org", _block_request_adapter)
    sess.mount("https://docs.python.org", _block_request_adapter)
    sess.mount("http://www.python.org", _block_request_adapter)
    sess.mount("https://www.python.org", _block_request_adapter)
    sess.mount("http://www.python.org/peps/", _block_request_adapter)
    sess.mount("https://www.python.org/peps/", _block_request_adapter)

    sess.mount("http://www.lfd.uci.edu/~gohlke/pythonlibs", _block_request_adapter)
    sess.mount("https://www.lfd.uci.edu/~gohlke/pythonlibs", _block_request_adapter)

    # Possibly useful, but needs to be used with care
    sess.mount("https://en.wikipedia.org", _block_request_adapter)
    sess.mount("https://packages.debian.org", _block_request_adapter)
    sess.mount("https://mirror.centos.org", _block_request_adapter)
    sess.mount("https://www.openhub.net", _block_request_adapter)  # 'html' package
    sess.mount("https://openhub.net", _block_request_adapter)
    sess.mount("https://www.ohloh.net", _block_request_adapter)
    sess.mount("https://web.archive.org", _block_request_adapter)
    sess.mount("https://badge.fury.io/py/", _block_request_adapter)
    sess.mount("https://www.linkedin.com", _block_request_adapter)
    sess.mount(
        "https://dealbook.nytimes.com", _block_request_adapter
    )  # github-issues-tools; leads to https://github.com/WP-API/WP-API
    sess.mount("http://hub.docker.com", _block_request_adapter)  # cauldron-notebook
    sess.mount("https://hub.docker.com", _block_request_adapter)
    sess.mount("http://docker.com", _block_request_adapter)  # wandb
    sess.mount("https://docker.com", _block_request_adapter)

    # google apps
    sess.mount("https://groups.google.com", _block_request_adapter)
    sess.mount("https://mail.google.com", _block_request_adapter)
    sess.mount("https://docs.google.com/forms/", _block_request_adapter)
    sess.mount(
        "https://docs.google.com/a/datastax.com/forms/", _block_request_adapter
    )  # cassandra-driver

    # Schema
    sess.mount(
        "http://schema.org/", _block_request_adapter
    )  # waiting/readthedocs, wordaxe
    sess.mount("https://schema.org/Article", _block_request_adapter)
    sess.mount("http://ogp.me/ns/", _block_request_adapter)
    sess.mount("https://ogp.me/ns/", _block_request_adapter)
    sess.mount("http://xmlns.com/", _block_request_adapter)
    sess.mount("https://xmlns.com/", _block_request_adapter)
    sess.mount("http://w3.org/TR/", _block_request_adapter)
    sess.mount("http://www.w3.org/TR/", _block_request_adapter)
    sess.mount("https://w3.org/TR/", _block_request_adapter)
    sess.mount("https://www.w3.org/TR/", _block_request_adapter)
    sess.mount("http://w3.org/199", _block_request_adapter)
    sess.mount("http://www.w3.org/199", _block_request_adapter)
    sess.mount("https://w3.org/199", _block_request_adapter)
    sess.mount("https://www.w3.org/199", _block_request_adapter)
    sess.mount("http://w3.org/200", _block_request_adapter)
    sess.mount("http://www.w3.org/200", _block_request_adapter)
    sess.mount("https://w3.org/200", _block_request_adapter)
    sess.mount("https://www.w3.org/200", _block_request_adapter)
    sess.mount("http://w3.org/201", _block_request_adapter)
    sess.mount("http://www.w3.org/201", _block_request_adapter)
    sess.mount("https://w3.org/201", _block_request_adapter)
    sess.mount("https://www.w3.org/201", _block_request_adapter)
    # irc
    sess.mount("http://webchat.freenode.net", _block_request_adapter)
    sess.mount("https://webchat.freenode.net", _block_request_adapter)
    sess.mount("http://irc.freenode.net", _block_request_adapter)
    sess.mount("https://irc.freenode.net", _block_request_adapter)
    sess.mount("http://gitter.im", _block_request_adapter)
    sess.mount("https://gitter.im", _block_request_adapter)
    # licenses
    sess.mount("http://apache.org/licenses/", _block_request_adapter)
    sess.mount("https://apache.org/licenses/", _block_request_adapter)
    sess.mount("http://www.apache.org/licenses/", _block_request_adapter)
    sess.mount("https://www.apache.org/licenses/", _block_request_adapter)
    sess.mount("https://www.creativecommons.org", _block_request_adapter)
    sess.mount("https://creativecommons.org", _block_request_adapter)
    sess.mount("https://wtfpl.net", _block_request_adapter)
    sess.mount("https://opensource.org/licenses/", _block_request_adapter)
    sess.mount("https://www.opensource.org/licenses/", _block_request_adapter)
    sess.mount("https://www.gnu.org/copyleft/", _block_request_adapter)
    sess.mount("https://www.gnu.org/licenses/", _block_request_adapter)
    sess.mount("https://www.gnu.org/philosophy/", _block_request_adapter)
    sess.mount("https://tldrlegal.com", _block_request_adapter)
    # pdf/papers
    sess.mount("https://arxiv.org/", _block_request_adapter)
    sess.mount("http://patft.uspto.gov/netacgi/", _block_request_adapter)  # very slow
    sess.mount("https://patft.uspto.gov/netacgi/", _block_request_adapter)  # BitVector
    sess.mount("http://joss.theoj.org/papers/", _block_request_adapter)  # umap-learn
    sess.mount("http://www.ietf.org/rfc/", _block_request_adapter)

    sess.mount(
        "https://gitter.im/allure-framework/allure-core", _block_request_adapter
    )  # allure-pytest
    sess.mount("https://tracpub.yaco.es", _block_request_adapter)
    sess.mount("http://tracpub.yaco.es", _block_request_adapter)

    # leads to docutils for many projects
    sess.mount("https://sphinx-doc.org", _block_request_adapter)
    sess.mount("http://sphinx-doc.org", _block_request_adapter)
    sess.mount("https://www.sphinx-doc.org", _block_request_adapter)
    sess.mount("http://www.sphinx-doc.org", _block_request_adapter)

    sess.mount("http://www.readthedocs.org", _block_request_adapter)
    sess.mount("https://www.readthedocs.org", _block_request_adapter)
    sess.mount("http://docs.readthedocs.org", _block_request_adapter)
    sess.mount("https://docs.readthedocs.org", _block_request_adapter)
    sess.mount("http://readthedocs.org/projects/", _block_request_adapter)
    sess.mount("https://readthedocs.org/projects/", _block_request_adapter)
    sess.mount("http://readthedocs.org/projects/", _block_request_adapter)
    sess.mount("https://readthedocs.org/projects/", _block_request_adapter)
    sess.mount(
        "https://fedoraproject.org/wiki/Infrastructure/Fedorahosted-retirement",
        _block_request_adapter,
    )
    # CI
    sess.mount("http://travis-ci.org", _block_request_adapter)
    sess.mount("http://travis-ci.com", _block_request_adapter)
    sess.mount("https://travis-ci.org", _block_request_adapter)
    sess.mount("https://travis-ci.com", _block_request_adapter)

    # More distracting than useful
    sess.mount("http://infrae.com", _block_request_adapter)
    sess.mount("https://infrae.com", _block_request_adapter)
    sess.mount(
        "http://peak.telecommunity.com/DevCenter/setuptools", _block_request_adapter
    )

    sess.mount("http://www.blueskyonmars.com", _block_request_adapter)
    sess.mount("https://www.blueskyonmars.com", _block_request_adapter)

    sess.mount("https://gitorious.org", NoVerifyHTTPAdapter())  # oset
    sess.mount("https://gisce.net", NoVerifyHTTPAdapter())  # electrical-calendar

    return sess


def get_file_cache_session(cache_name):
    cache_path = cache_subdir(cache_name)

    https_exceptions = {
        "code.welldev.org",  # pypi oauth_provider; nothing on https; http is parked
        "fedmsg.com",  # https redirects to http, causing too-many-redirects
        "farmdev.com",  # SSL_ERROR_RX_RECORD_TOO_LONG
        "abyz.me.uk",  # expired pigpio
        "api.arduino.cc",
        "cdn.arduino.cc",  # rpi-gpio
        "lamb.cc",  # script
        # https very slow:
        "pybrary.net",
        "fabfile.org",  # port 80 also timing out
        "paramiko.org",  # port 80 also timing out
        "pyinvoke.org",  # port 80 also timing out
        "www.freecell.org",
        "www.dnspython.org",
        "stutzbachenterprises.com",
        "blockdiag.com",
        "welbornprod.com",  # :80 redirects to github
        "code.krypto.org",  # hashlib
        "intelligent-systems.altran.com",  # spark-parser
        "dalkescientific.com",  # PyRSS2Gen
        "luispedro.org",  # pymorph
        "dir.gmane.org",  # dm.xmlsec-binding
        "bibframe.org",  # pybibframe
        "galaxyproject.org",  # galaxy
        # nothing on https:
        "www.mypy-lang.org",
        "www.decida.org",
        "www.alcyone.com",
        "www.aaronsw.com",
        "whitemans.ca",
        "www.roundup-tracker.org",
        "code.mibian.net",
        "peak.telecommunity.com",
        "www.ruffus.org.uk",
        "kaino.kotus.fi",
        "wordlist.aspell.net",
        "www.python-excel.org",
        "sparrow.telecommunity.com",
        "blog.pediapress.com",
        "eh.org",
        "www.linuxrising.org",  # pygtk
        "www.4guysfromrolla.com",
        "svn.eby-sarna.com",
        "azure.archive.ubuntu.com",  # blurb
        "www.should-dsl.info",  # should_dsl
        "cvs.bigasterisk.com",  # pyalsa
        "decida.org",  # DeCiDa
        "antirez.com",  # redlock-py
        "docs.h2o.ai",  # https://github.com/h2oai/sparkling-water/issues/1953
        "www.theblobshop.com",  # https://github.com/jmyounker/vers/issues/1
        # wrong cert
        "pybrain.org",  # cert for *.kasserver.com, kasserver.com (unrelated)
        "www.unnotebook.com",  # cloudfront.net, *.cloudfront.net
        "download.ros.org",  # *.osuosl.org
        "fimi.ua.ac.be",  # pyfim
        "kassiopeia.juls.savba.sk",  # unicode; errkorp.juls.savba.sk
        "www.hackersdelight.org",  # *.ipower.com
        "download.qt-project.org",  # *.qt.io
        "occiput.scienceontheweb.net",  # petlink
        "morpho.aalto.fi",  # Morfessor
        "www.anderoid.com",  # BitVector
        "urlgrabber.baseurl.org",  # beaverbarcamp.org, www.beaverbarcamp.org
        "celljam.net",  # *.webfaction.com, webfaction.com
        "www.acooke.org",  # *.webfaction.com, webfaction.com
        "timgolden.me.uk",  # *.webfaction.com, webfaction.com
        "mpmath.org",  # *.webfaction.com, webfaction.com
        "www.myhdl.org",  # *.webfaction.com, webfaction.com
        "science.webhostinggeeks.com",  # webhostinggeeks.com, www.webhostinggeeks.com
        "sphinx.pocoo.org",  # pocoo.org, www.pocoo.org
        "lists.baruwa.org",  # *.baruwa.net, baruwa.net
        "yum.baseurl.org",  # beaverbarcamp.org
        "python-tmx.nongnu.org",  # www.gnu.org and friends
        "cthedot.de",  # *.mivitec.net
        "www.acoular.org",  # *.webgo24.de
        "bibfra.me",  # *.library.link
        "www.grantjenks.com",  # https://github.com/grantjenks/python-diskcache/issues/141
        "www.nomadblue.com",  # *.herokuapp.com
        "www.likit.lt",  # ebcdic, *.serveriai.lt
        "www.jsnp.net",  # magic
        "demo.drfdocs.com",  # cert github
        "spockframework.org",  # github
        "schemaform.io",  # github
        "octopress.org",  # github
        "augeas.net",  # github
        "blog.myrhy.me",  # github
        "mypaint.org",  # github, pygtyk
        "www.cheetahtemplate.org",  # github
        "www.keyczar.org",  # redirects to github, hosted by shortener.secureserver.net
        "nltk.org",  # redirects to docs site, hosted by shortener.secureserver.net
        "pynag.org",
        "trivas.pl",
        "psychtoolbox.org",
        "app.datarobot.com",
        "modestmaps.com",  # github
        #'geopandas.org',  # cert github fails https://github.com/geopandas/geopandas/issues/1287
        #'pysolar.org',  # cert github fails
        "www.pyqtgraph.org",  # cert github fails
        "www.buildout.org",
        "docs.buildout.org",  # *.readthedocs.io, readthedocs.io
        "babel.pocoo.org",  # *.readthedocs.io, readthedocs.io
        "www.zodb.org",  # *.readthedocs.io, readthedocs.io
        "www.pygal.org",
        "cloudbridge.cloudve.org",  # readthedocs
        "www.getlfs.com",  # cert wrong host
        "demo.getlfs.com",  # www.holz-hoehne.de
        "docs.getlfs.com",  # rtd
        "git.tremily.us",  # blog.tremily.us
        "gfxmonk.net",  # https://github.com/timbertson/termstyle/issues/10
        # self signed
        "www.larryhastings.com",  # self signed
        "github-api.hive.pt",  # self signed
        "www.goof.com",  # BitVector
        # www.modwsgi.org special case redirects to rtd
        "konlpy.org",  # *.readthedocs.org, readthedocs.org
        "deeplearning.net",  # expired
        "www.beian.gov.cn",  # expired
        "www.functx.com",
        "lists2.idyll.org",  # bad cert
        # ID tracker
        "p01.notifa.info",
        # indirect
        "api.del.icio.us",  # pyalsa; hostname 'api.del.icio.us' doesn't match either of '*.pinboard.in', 'pinboard.in'
        # ok in Firefox, but cert verify failed
        "www.dabeaz.com",  # This server supports TLS 1.0 and TLS 1.1.  This server's certificate chain is incomplete.
        "html5up.net",
        "pages.cpsc.ucalgary.ca",  # https://www.ssllabs.com/ssltest/analyze.html?d=pages.cpsc.ucalgary.ca&latest
        "www.thevirtualbrain.org",
        "web.resource.org",
        "zesty.ca",
    }

    base = os.path.dirname(__file__)

    session = requests.Session()
    session = CacheControl(
        session,
        cache=FileCache(cache_path),
        controller_class=MoreCodesCacheController,
        adapter_class=ForceTimeoutHTTPAdapter,
        cacheable_methods=("GET"),  # https://github.com/ionrock/cachecontrol/issues/216
        heuristic=IgnoreVaryExpiresAfter(days=5),
        blocklist=os.path.join(base, "park_providers.txt"),
        https_exceptions=https_exceptions,
    )
    session.max_redirects = MAX_REDIRECTS

    if cache_name == "web":
        session.mount("http://pypi.org", _block_request_adapter)
        session.mount("https://pypi.org", _block_request_adapter)
        session.mount("http://pypi.python.org", _block_request_adapter)
        session.mount("https://pypi.python.org", _block_request_adapter)
        session.mount("http://cheeseshop.python.org", _block_request_adapter)
        session.mount("https://cheeseshop.python.org", _block_request_adapter)
        session.mount("http://UNKNOWN", _block_request_adapter)
        session.mount("https://UNKNOWN", _block_request_adapter)

    return session


get_file_cache = get_file_cache_session



from unittest_expander import expand, foreach

from pypidb._pagure import _pagure_io
from tests.utils import _TestBase, web_session

pagure_io = ["copr", "copr-common", "python-daemon", "rustcfg", "wikitcms"]


@expand
class TestPagureIO(_TestBase):
    @foreach(pagure_io)
    def test_package(self, name):
        url = self.converter.get_vcs(name)
        self.assertIsNotNone(url)
        self.assertIn("//pagure.io/", url)
        r = web_session.head(url)
        r.raise_for_status()

    def test_daemon(self):
        url = self.converter.get_vcs("python-daemon")
        self.assertInsensitiveEqual(url, "https://pagure.io/python-daemon")

    def test_daemon_issues(self):
        url = _pagure_io("https://pagure.io/python-daemon/issues")
        self.assertInsensitiveEqual(url, "https://pagure.io/python-daemon")

    def test_invalid(self):
        url = _pagure_io("https://pagure.io/invalid")
        self.assertIsNone(url)

    def test_wikitcms(self):
        url = self.converter.get_vcs("wikitcms")
        self.assertEqual(url, "https://pagure.io/fedora-qa/python-wikitcms")

    def test_meta_repo(self):
        url = _pagure_io("https://pagure.io/fedora-qa/foo/bar")
        self.assertEqual(url, "https://pagure.io/fedora-qa")



import unittest

from unittest_expander import expand, foreach

from tests.data import expected
from tests.utils import _stdlib, _TestBase


def _fetch_names():
    for name in _stdlib:
        # Already tested in separate class
        if name in expected:
            continue

        yield name


@expand
class TestStdlibNames(_TestBase):

    names = sorted(_fetch_names())
    _ignore_invalid = True
    _allow_missing = True
    expected_failures = [
        "ast",
        "chunk",
        "dis",
        "formatter",
        "mailbox",
        "modulefinder",
        "resource",
        "secrets",
        "token",
        "turtle",
        "wave",
        # has url
        "numbers",
        "parser",
        "select",
        "shelve",
        "signal",
        # 'time' has no urls or artifacts
        "trace",
    ]
    # numbers and calendar and trace and signal and shelve and select and parser have urls
    #   calendar republished to https://pypi.org/project/python-calendrical/
    # numbers, select and signal point to github repos which are 404
    #   select was renamed to https://github.com/Jaymon/que
    # time has no urls or artifacts
    # trace links to http://billionuploads.com/ka79h2t4jpi1 which looks dodgy but is 404 atm

    @foreach(names)
    def test_package(self, name):
        self._test_names([name])


@expand
class TestFailedStdlibNames(_TestBase):

    expected_failures = []

    @foreach(TestStdlibNames.expected_failures)
    @unittest.expectedFailure
    def test_package(self, name):
        self._test_names([name])
